#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import json
import unittest

from gemmi import cif
from common import full_path

def nonempty_lines(s):
    return [line for line in s.splitlines() if line and line[0] != '#']

class TestCifAsJson(unittest.TestCase):
    def test_misc(self):
        basename = full_path('misc')
        cif_doc = cif.read_file(basename + '.cif')
        json_str = cif_doc.as_json()
        json_from_cif = json.loads(json_str)
        with io.open(basename + '.json', encoding='utf-8') as f:
            reference_json = json.load(f)
        self.assertEqual(json_from_cif, reference_json)

    def test_cif_as_string(self):
        with open(full_path('misc.cif'), 'rb') as f:
            cif_orig = f.read().decode('utf-8').replace('\r\n', '\n')
        cif_doc = cif.read_string(cif_orig)
        formatting_changes = {
            '# comment\n': '',
            '_x _y _z': '_x\n_y\n_z',
            '_b 2 3 4': '_b\n2\n3\n4',
            'pdbx_details    ': 'pdbx_details ',
        }
        for k, v in formatting_changes.items():
            cif_orig = cif_orig.replace(k, v)
        cif_out = cif_doc.as_string()
        self.assertListEqual(nonempty_lines(cif_orig), nonempty_lines(cif_out))

    def test_options(self):
        doc = cif.read_string('data_OnE loop_ _TwO 2 _ZeRo 0')
        j1 = json.loads(doc.as_json())
        self.assertEqual(j1, {'one': {'_two': [2], '_zero': 0}})
        j2 = json.loads(doc.as_json(lowercase_names=False))
        self.assertEqual(j2, {'OnE': {'_TwO': [2], '_ZeRo': 0}})

    def test_json_number_normalization(self):
        # JSON requires a digit before decimal point, CIF allows -.99 and .04
        doc = cif.read_string('data_test _a -.99 _b .04 _c -.04 _d +.5')
        json_str = doc.as_json()
        # The JSON must be parseable by standard parser
        parsed = json.loads(json_str)
        self.assertEqual(parsed, {'test': {'_a': -0.99, '_b': 0.04,
                                           '_c': -0.04, '_d': 0.5}})

class TestMmjson(unittest.TestCase):
    def test_read_1pfe(self):
        path = full_path('1pfe.json')
        doc = cif.read_mmjson(path)
        self.assertEqual(doc[0].find_value('_entry.id'), '1PFE')
        output_json = json.loads(doc.as_json(mmjson=True))
        with io.open(path, encoding='utf-8') as f:
            input_json = json.load(f)
        # We don't preserve the type information from json and some strings
        # are written back as numbers.
        # Here we convert them to strings before comparison.
        for category_name, category in output_json["data_1PFE"].items():
            ref_category = input_json["data_1PFE"][category_name]
            for tag, values in category.items():
                ref_values = ref_category[tag]
                if ref_values != values:
                    for n, value in enumerate(values):
                        if value is not None:
                            values[n] = str(value)
        self.assertEqual(output_json, input_json)

    def test_read_3wup(self):
        path = full_path('3wup.json.gz')
        doc = cif.read_mmjson(path)
        self.assertEqual(doc[0].find_value('_entry.id'), '3WUP')

    def test_read_bytes(self):
        doc = cif.read_mmjson_string(b'{"data_1PFE":{"entry":{"id":["1ABC"]}}}')
        self.assertEqual(doc[0].find_value('_entry.id'), '1ABC')

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
import sys
from gemmi import cif

for path in sys.argv[1:]:
    try:
        doc = cif.read_file(path)  # copy all the data from mmCIF file
        for block in doc:  # iterate over blocks
            for cc in block.find("_chem_comp.", ["id", "formula_weight"]):
                print(cc[0], "weights", cc[1])
    except Exception as e:
        print("Oops. %s" % e)



#!/usr/bin/env python3

# Plots the data from rama_gather.py

import sys
import matplotlib.pyplot as plt
from matplotlib.ticker import MultipleLocator

def plot(data_file, label, output=None):
    x, y = [], []
    with open(data_file) as f:
        for line in f:
            phi, psi = line.split()
            if phi != 'nan' and psi != 'nan':
                x.append(float(phi))
                y.append(float(psi))
    print('Plotting %d points for %s' % (len(x), label))

    plt.figure(figsize=(5.5, 5.5))
    plt.title('%s, %d points.' % (label, len(x)), fontsize=12)
    plt.xlim([-180, 180])
    plt.ylim([-180, 180])
    ax = plt.gca()
    ax.xaxis.set_major_locator(MultipleLocator(60))
    ax.yaxis.set_major_locator(MultipleLocator(60))
    plt.xlabel(r'$\phi$', fontsize=14)
    plt.ylabel(r'$\psi$', fontsize=14, labelpad=0)
    plt.grid(color='#AAAAAA', linestyle='--')
    plt.hexbin(x, y, gridsize=2*180, bins='log', cmap='Blues')
    if output:
        plt.savefig(output, dpi=300)  # dpi=70 for small images in docs
    else:
        plt.show()

for aa in sys.argv[1:]:
    plot('ramas/%s.tsv' % aa, aa)
    #plot('ramas/%s.tsv' % aa, aa, 'ramas/%s.png' % aa)



# -*- coding: utf-8 -*-

# -- General configuration ------------------------------------------------

# while we use Sphinx 8+, old version suffices to run doctests
needs_sphinx = '5.3.0'

extensions = ['sphinx.ext.doctest', 'sphinx_inline_tabs']

templates_path = ['_templates']

master_doc = 'index'

project = u'Gemmi'
copyright = u'Global Phasing Ltd'
author = u'Marcin Wojdyr'

with open('../include/gemmi/version.hpp') as _f:
    for _line in _f:
        if _line.startswith('#define GEMMI_VERSION '):
            version = _line.split()[2].strip('"')
release = version

# now sure if we'll use headers.rst again, disable it for now
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'headers.rst']
pygments_style = 'sphinx'
todo_include_todos = False
highlight_language = 'cpp'
default_role = 'literal'


# -- Options for HTML output ----------------------------------------------

html_theme = 'furo'
html_theme_options = {
    "source_repository": "https://github.com/project-gemmi/gemmi/",
    "source_branch": "master",
    "source_directory": "docs/",
}
html_static_path = ['_static']
html_css_files = ['custom.css']

# Edit link can be also used to see the source
html_show_sourcelink = False
html_copy_source = False

def setup(app):
    app.connect("builder-inited", monkey_patching_furo)

def monkey_patching_furo(app):
    if app.builder.name != 'html':
        return

    import furo
    from typing import Dict, Any
    def _compute_navigation_tree(context: Dict[str, Any]) -> str:
        # The navigation tree, generated from the sphinx-provided ToC tree.
        if "toctree" in context:
            toctree = context["toctree"]
            toctree_html = toctree(
                collapse=False,
                titles_only=False,
                maxdepth=2,
                includehidden=True,
            )
        else:
            toctree_html = ""
        return furo.get_navigation_tree(toctree_html)

    furo._compute_navigation_tree = _compute_navigation_tree

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # 'papersize': 'letterpaper',
    # 'pointsize': '10pt',
    # 'preamble': '',
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'Gemmi.tex', u'Gemmi Documentation',
     u'Marcin Wojdyr', 'manual'),
]

doctest_global_setup = '''
import os
import sys
disabled_features = []
try:
    import numpy
    if numpy.__version__ >= '2':
        numpy.set_printoptions(legacy='1.25')
    numpy.set_printoptions(threshold=5)
except ImportError:
    disabled_features.append('NumPy')
    numpy = None
try:
    import pandas
except ImportError:
    disabled_features.append('pandas')
    pandas = None
try:
    import networkx
except ImportError:
    disabled_features.append('networkx')
    networkx = None
try:
    import pynauty
except ImportError:
    disabled_features.append('pynauty')
    pynauty = None
ccp4_path = os.getenv('CCP4')
if ccp4_path:
    mdm2_unmerged_mtz_path = (ccp4_path + '/lib/python3.9/site-packages/'
                              + 'ccp4i2/demo_data/mdm2/mdm2_unmerged.mtz')
    if not os.path.isfile(mdm2_unmerged_mtz_path):
        ccp4_path = None
if ccp4_path is None:
    disabled_features.append('$CCP4')

import gemmi
gemmi.set_leak_warnings(False)
'''



#!/usr/bin/env python
# Check amino-acid frequency in the PDB database (or it's subset)
# by reading meta-data from mmCIF files.

import sys
from collections import Counter
from gemmi import cif, CifWalk

totals = Counter()
for arg in sys.argv[1:]:
    for path in CifWalk(arg, try_pdbid='M'):
        # read file (uncompressing on the fly) and get the only block
        block = cif.read(path).sole_block()
        # find table with the sequence
        seq = block.find('_entity_poly_seq.', ['entity_id', 'mon_id'])
        # convert table with chain types (protein/DNA/RNA) to dict
        entity_types = dict(block.find('_entity_poly.', ['entity_id', 'type']))
        # and count these monomers that correspond to a protein chain
        aa_counter = Counter(row.str(1) for row in seq
                             if 'polypeptide' in entity_types[row.str(0)])
        totals += aa_counter
        # print residue counts for each file
        print(block.name, *(f'{m}:{c}' for (m, c) in aa_counter.most_common()))
# finally, print the total counts as percentages
f = 100.0 / sum(totals.values())
print('TOTAL', *(f'{m}:{c*f:.2f}%' for (m, c) in totals.most_common(20)))



#!/usr/bin/env python
# Check the space-groups.txt from Open Babel
# Usage: ob_spacegroups.py /path/to/space-groups.txt

from __future__ import print_function
import sys
import gemmi

space_groups_txt = sys.argv[1]
chunks = open(space_groups_txt).read().split('\n\n')
def parse_chunk(lines):
    return {'number': int(lines[0]),
            'hall': lines[1].strip(),
            'xhm': lines[2].strip(),
            'symops': [gemmi.Op(line) for line in lines[3:]]}
data = [parse_chunk(c.splitlines()) for c in chunks if c]
print(len(data), 'items')
for d in data:
    ops = gemmi.symops_from_hall(d['hall'])
    assert len(ops.sym_ops) * len(ops.cen_ops) == len(d['symops'])
    given = set(d['symops'])
    generated = set(ops)
    if given != generated:
        print(d['hall'])
        print('common:',  '  '.join(x.triplet() for x in given & generated))
        print('given:    ', '  '.join(x.triplet() for x in given - generated))
        print('generated:', '  '.join(x.triplet() for x in generated - given))
for d in data:
    names = d['xhm'].split(',')
    if all(gemmi.find_spacegroup_by_name(name) is None for name in names):
        print('not in gemmi:', names)



# Gathers data for Ramachandran plot that can be plotted with rama_plot.py.
# See the section about torsion angles in documentation.

import sys
from math import degrees
import gemmi

ramas = {aa: [] for aa in [
    'LEU', 'ALA', 'GLY', 'VAL', 'GLU', 'SER', 'LYS', 'ASP', 'THR', 'ILE',
    'ARG', 'PRO', 'ASN', 'PHE', 'GLN', 'TYR', 'HIS', 'MET', 'CYS', 'TRP']}

for path in gemmi.CoorFileWalk(sys.argv[1]):
    st = gemmi.read_structure(path)
    if 0.1 < st.resolution < 1.5:
        model = st[0]
        for chain in model:
            for res in chain.get_polymer():
                # previous_residue() and next_residue() return previous/next
                # residue only if the residues are bonded. Otherwise -- None.
                prev_res = chain.previous_residue(res)
                next_res = chain.next_residue(res)
                if prev_res and next_res and next_res.name != 'PRO':
                    v = gemmi.calculate_phi_psi(prev_res, res, next_res)
                    try:
                        ramas[res.name].append(v)
                    except KeyError:
                        pass

# Write data to files
for aa, data in ramas.items():
    with open('ramas/' + aa + '.tsv', 'w') as f:
        for phi, psi in data:
            f.write(f'{degrees(phi):.4f}\t{degrees(psi):.4f}\n')



#!/usr/bin/env python
"Find PDB entries with more than 50,000 anisotropic B-factors."

import sys
from gemmi import cif, CifWalk

for path in CifWalk(sys.argv[1]):
    block = cif.read(path).sole_block()
    anis = block.find_values("_atom_site_anisotrop.id")
    if len(anis) > 50000:
        print(block.name, len(anis))



#!/usr/bin/env python3

import sys
import zipfile

if len(sys.argv) < 3:
    sys.exit("specify whl files")

def get_names(zip_obj):
    return sorted(x for x in zip_obj.namelist() if not x.endswith('/'))

with zipfile.ZipFile(sys.argv[1], 'r') as z:
    print('Reference:', sys.argv[1])
    ref_lst = set(get_names(z))
for arg in sys.argv[2:]:
    print('Comparing with', arg)
    with zipfile.ZipFile(arg, 'r') as z:
        lst = get_names(z)
    for name in ref_lst:
        if name in lst:
            lst.remove(name)
        else:
            print('    - ', name)
    for name in lst:
        print('    + ', name)



#!/usr/bin/env python3

# Check for new versions of third-party libraries

from __future__ import print_function
import datetime
import json
from urllib.request import urlopen

TAGGED_REPOS = {
    'wjakob/nanobind': 'v2.6.1',
    'scikit-build/scikit-build-core': 'v0.11.5',
    'taocpp/PEGTL': '2.8.3',
    'cxong/tinydir': '1.2.6',
    'fastfloat/fast_float': 'v8.2.3',
    'madler/zlib': 'v1.3.2',
    'LLNL/shroud': 'v0.14.0',
}

NOT_TAGGED_REPOS = {
    'chadaustin/sajson': 'include/sajson.h',
    'sgorsten/linalg': 'linalg.h',
    'nothings/stb': 'stb_sprintf.h',
}

def load_json(url):
    with urlopen(url) as response:
        return json.load(response)

def check_tags():
    for repo, version in TAGGED_REPOS.items():
        try:
            url = 'https://api.github.com/repos/%s/releases/latest' % repo
            data = load_json(url)
            latest_tag = data['tag_name']
        except IOError:
            data = load_json('https://api.github.com/repos/%s/tags' % repo)
            latest_tag = data[0]['name']
        mark = ('   !!!' if version != latest_tag else '')
        print('%-18s %10s %10s%s' % (repo[-18:], version, latest_tag, mark))

def check_recent_commits():
    since = datetime.datetime.now() - datetime.timedelta(days=90)
    for repo, filename in NOT_TAGGED_REPOS.items():
        url = 'https://api.github.com/repos/%s/commits?path=%s&since=%s' % (
              repo, filename, since.isoformat())
        data = load_json(url)
        info = '-'
        if data:
            info = data[0]['commit']['committer']['date'][:10]
        print('%-18s  modified: %s' % (filename.split('/')[-1], info))

try:
    check_tags()
    check_recent_commits()
except IOError as e:
    print(e)



#!/usr/bin/env python

import sys
import numpy
import gemmi

def maskdiff(path1, path2):
    mask1 = gemmi.read_ccp4_mask(path1, setup=True)
    arr1 = mask1.grid.array
    mask2 = gemmi.read_ccp4_mask(path2, setup=True)
    arr2 = mask2.grid.array
    print("Size: %d x %d x %d  and  %d x %d x %d" % (arr1.shape + arr2.shape))
    if arr1.shape != arr2.shape:
        sys.exit("Different sizes. Exiting.")
    t = 2 * (arr1 != 0) + (arr2 != 0)
    for (a, b) in [(0, 0), (1, 1), (0, 1), (1, 0)]:
        n = numpy.count_nonzero(t == 2*a+b)
        print('%d-%d %12d %6.2f%%' % (a, b, n, 100.*n/arr1.size))
    print('total %10d' % arr1.size)

if __name__ == '__main__':
    if len(sys.argv) != 3:
        sys.exit("Usage: maskdiff.py map1 map2")
    maskdiff(sys.argv[1], sys.argv[2])



#!/usr/bin/env python
# Reads a mask, reads a coordinate file, and checks how the mask compares
# with a bulk solvent mask for the coordinate file.

# To check if we produce identical maps as Refmac, use tools/compare_mask.sh.
#
# To check if we produce identical maps as cctbx, do:
#   cctbx.python -m mmtbx.command_line.mask file.pdb
#   maskcheck.py mask.ccp4 file.pdb

'Usage: maskcheck.py [-v] MAP_FILE COORDINATE_FILE [OUTPUT_MAP]'

import os
import sys
import numpy
import gemmi

def maskcheck(mask_path, coor_path, output_diff_map=None, verbose=False):
    # read mask
    mask = gemmi.read_ccp4_mask(mask_path, setup=True)
    grid = mask.grid

    # read coordinates
    st = gemmi.read_structure(coor_path)

    # check if unit cell and symmetry are the same
    if not grid.unit_cell.approx(st.cell, epsilon=1e-4):
        print('Cell parameters differ. PDB:', st.cell.parameters)
        print('                       Mask:', grid.unit_cell.parameters)
    sg = st.find_spacegroup()
    if grid.spacegroup != sg:
        print('Space groups differ:', grid.spacegroup.xhm(), 'vs', sg.xhm())

    # prepare a mask with the same size
    if 'REFMAC' in os.environ:
        print('Generating REFMAC-compatible mask ...')
        masker = gemmi.SolventMasker(gemmi.AtomicRadiiSet.Refmac)
        masker.island_min_volume = 8 * 2.8**3
    else:
        print('Generating CCTBX-compatible mask ...')
        masker = gemmi.SolventMasker(gemmi.AtomicRadiiSet.Cctbx)
    grid2 = gemmi.Int8Grid()
    grid2.copy_metadata_from(grid)
    masker.put_mask_on_int8_grid(grid2, st[0])

    compare_mask_arrays(grid, grid2, verbose)
    if verbose:
        print_nearby_atoms(st, grid, grid2)
    if output_diff_map:
        write_diff_map(grid, grid2, output_diff_map)


def compare_mask_arrays(grid1, grid2, verbose):
    arr1 = grid1.array
    arr2 = grid2.array
    if arr1.shape != arr2.shape:
        sys.exit('Different grid sizes %s and %s. Exiting.' %
                 (arr1.shape, arr2.shape))
    print('Size: %d x %d x %d,' % arr1.shape, 'total', arr1.size, 'points')
    if (arr1 == arr2).all():
        print("Identical.")
        if not verbose:
            return
    t = 2 * (arr1 != 0) + (arr2 != 0)
    print('File-Gemmi Count Fraction')
    for (a, b) in [(0, 0), (1, 1), (0, 1), (1, 0)]:
        n = numpy.count_nonzero(t == 2*a+b)
        print('%d-%d %12d %6.2f%%' % (a, b, n, 100.*n/arr1.size))

# Print nearby atom for each differing point
def print_nearby_atoms(st, grid1, grid2):
    ns = gemmi.NeighborSearch(st[0], st.cell, 4).populate()
    diff_grid = get_diff_grid(grid1, grid2)
    for negate in (True, False):
        blobs = gemmi.find_blobs_by_flood_fill(diff_grid, cutoff=0,
                                               min_volume=0, min_score=0,
                                               negate=negate)
        blobs.sort(key=lambda a: a.volume, reverse=True)
        print('\n%s %d blobs' % ((negate and '0-1' or '1-0'), len(blobs)))
        for blob in blobs:
            cra = ns.find_nearest_atom(blob.centroid).to_cra(st[0])
            print(f'    {blob.volume:.1f} A^3 near {cra}')

def get_diff_grid(grid1, grid2):
    arr = (grid1.array - grid2.array).astype(dtype=numpy.float32)
    return gemmi.FloatGrid(arr, grid1.unit_cell, grid1.spacegroup)

def write_diff_map(grid1, grid2, output_diff_map):
    map_out = gemmi.Ccp4Map()
    map_out.grid = get_diff_grid(grid1, grid2)
    map_out.update_ccp4_header()
    map_out.write_ccp4_map(output_diff_map)

def main():
    args = sys.argv[1:]
    verbose = ('-v' in args)  # pylint: disable=superfluous-parens
    if verbose:
        args.remove('-v')
    if len(args) not in (2, 3):
        sys.exit(__doc__)
    maskcheck(*args, verbose=verbose)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# IF YOU ADD OR REMOVE LINES, ADJUST :lines: in docs/hkl.rst

import pandas
from matplotlib import pyplot
import gemmi

MTZ_PATH = 'tests/5wkd_phases.mtz.gz'
SFCIF_PATH = 'tests/r5wkdsf.ent'

# make DataFrame from MTZ file
mtz = gemmi.read_mtz_file(MTZ_PATH)
mtz_df = pandas.DataFrame(data=mtz.array, columns=mtz.column_labels())
# (optional) store Miller indices as integers
mtz_df = mtz_df.astype({label: 'int32' for label in 'HKL'})

# make DataFrame from mmCIF file
cif_doc = gemmi.cif.read(SFCIF_PATH)
rblock = gemmi.as_refln_blocks(cif_doc)[0]
cif_df = pandas.DataFrame(data=rblock.make_miller_array(),
                          columns=['H','K','L'])
cif_df['F_meas_au'] = rblock.make_float_array('F_meas_au')
cif_df['d'] = rblock.make_d_array()

# merge DataFrames
df = pandas.merge(mtz_df, cif_df, on=['H', 'K', 'L'])


# plot FP from MTZ as a function of F_meas_au from mmCIF
pyplot.rc('font', size=8)
pyplot.figure(figsize=(2, 2))
pyplot.scatter(x=df['F_meas_au'], y=df['FP'],
               marker=',', s=1, linewidths=0)
pyplot.xlim(xmin=0)
pyplot.ylim(ymin=0)
pyplot.show()


# plot the ratio FP : F_meas_au as a function of 1/d
pyplot.figure(figsize=(3, 3))
pyplot.scatter(x=1/df['d'],
               y=df['FP']/df['F_meas_au'],
               c=df['K'],  # color by index k
               marker='.', s=16, linewidths=0)
pyplot.xlim(xmin=0)
pyplot.show()



#!/usr/bin/env python

import unittest
import os
import gemmi

class TestMisc(unittest.TestCase):
    def test_pdb_code(self):
        self.assertTrue(gemmi.is_pdb_code('9ALA'))
        self.assertFalse(gemmi.is_pdb_code('BAD1'))
        pdb_dir = os.getenv('PDB_DIR')
        if not pdb_dir:
            return
        cif_4xyz = pdb_dir + '/structures/divided/mmCIF/xy/4xyz.cif.gz'
        self.assertEqual(gemmi.expand_pdb_code_to_path('4XYZ', 'M'), cif_4xyz)
        self.assertEqual(gemmi.expand_if_pdb_code('4XYZ'), cif_4xyz)
        self.assertEqual(gemmi.expand_if_pdb_code('4XYZ', filetype='P'),
                         pdb_dir + '/structures/divided/pdb/xy/pdb4xyz.ent.gz')

    def test_selections(self):
        def selstr(s):
            s2 = gemmi.Selection(s).str()
            s3 = gemmi.Selection(s2).str()
            self.assertEqual(s2, s3)
            return s2
        self.assertEqual(selstr('/1'), '/1/*//')
        self.assertEqual(selstr('CA:*'), '//*//CA')
        self.assertEqual(selstr('CA[C]:'), '//*//CA[C]:')
        self.assertEqual(selstr('CA[C]:A'), '//*//CA[C]:A')
        self.assertEqual(selstr('A;b>20'), '//A//;b>20')
        self.assertEqual(selstr('/1;q>0.5'), '/1/*//;q>0.5')
        self.assertEqual(selstr('[C];b<20'), '//*//[C];b<20')
        self.assertEqual(selstr(':;b=20'), '//*//:;b=20')
        self.assertEqual(selstr(':B'), '//*//:B')
        self.assertEqual(selstr(':B;q=0'), '//*//:B;q=0')
        self.assertEqual(selstr('A//CA'), '//A//CA')
        self.assertEqual(selstr('(ALA)'), '//*/(ALA)/')
        self.assertEqual(selstr('(ALA,GLY)'), '//*/(ALA,GLY)/')
        self.assertEqual(selstr('15-55'), '//*/15.-55./')
        self.assertEqual(selstr('15C-55B'), '//*/15.C-55.B/')
        self.assertEqual(selstr('15.C-55.B'), '//*/15.C-55.B/')
        self.assertEqual(selstr('///'), '////')
        self.assertEqual(selstr('[Cu]'), '//*//[Cu]')
        self.assertEqual(selstr('[Mg,O,X]'), '//*//[X,O,Mg]')
        self.assertEqual(selstr('[!Xe]'), '//*//[!Xe]')
        self.assertEqual(selstr('[!H,D]'), '//*//[!H,D]')
        self.assertEqual(selstr(';q=0'), '//*//;q=0')
        self.assertEqual(selstr('(ALA);polymer;b<30;q>0'),
                         '//*/(ALA)/;polymer;b<30;q>0')
        self.assertEqual(selstr('[!H,D];!polymer,solvent'),
                         '//*//[!H,D];!polymer,solvent')
        self.assertEqual(selstr('A/33.-120.A/[C]'), '//A/33.-120.A/[C]')
        self.assertEqual(selstr('B/7'), '//B/7./')
        self.assertEqual(selstr('/1/C/-72(DA)'), '/1/C/-72.(DA)/')
        self.assertEqual(selstr('C/-72--30'), '//C/-72.--30./')
        self.assertEqual(selstr('C/-72-2'), '//C/-72.-2./')
        self.assertEqual(selstr('[metals]'), selstr('[!nonmetals]'))
        self.assertEqual(selstr('[!metals]'), selstr('[nonmetals]'))
        self.assertTrue('[X,H,B' in selstr('[!metals,He]'))
        self.assertTrue('[!X,H,B' in selstr('[metals,He]'))

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import unittest
import gemmi
from common import full_path

# In 5a11 applying NCS causes atom clashing
FRAGMENT_5A11 = """\
CRYST1   48.367   89.613   83.842  90.00 101.08  90.00 P 1 21 1      4          
MTRIX1   1 -0.999980  0.006670  0.000050       13.74419                         
MTRIX2   1  0.006260  0.941760 -0.336220       35.56956                         
MTRIX3   1 -0.002290 -0.336210 -0.941780      205.34927                         
ATOM    256  SG  CYS A  37      -1.002 -31.125  88.394  1.00 14.38           S  
ATOM   2969  SG  CYS B  37      14.582 -23.455 132.554  1.00 18.14           S  
"""  # noqa: W291 - trailing whitespace

# In 1gtv two different chains (with partial occupancy) are exactly in the
# same place
FRAGMENT_1GTV = """\
CRYST1   76.353   76.353  134.815  90.00  90.00 120.00 P 65 2 2     24
ATOM    635  SG  CYS A  85      42.948   6.483  17.913  0.48 23.86           S
ATOM   2293  SG  CYS B  85      42.948   6.483  17.913  0.52 23.86           S
"""

class TestNeighborSearch(unittest.TestCase):
    def test_5a11(self, use_populate=True):
        st = gemmi.read_pdb_string(FRAGMENT_5A11)
        a1 = st[0].sole_residue('A', gemmi.SeqId(37, ' '))[0]
        ns = gemmi.NeighborSearch(st[0], st.cell, 5)
        if use_populate:
            ns.populate()
        else:
            for n_ch, chain in enumerate(st[0]):
                for n_res, res in enumerate(chain):
                    for n_atom, atom in enumerate(res):
                        ns.add_atom(atom, n_ch, n_res, n_atom)
        marks = ns.find_atoms(a1.pos, a1.altloc, radius=3)
        m1, m2 = sorted(marks, key=lambda m: ns.dist(a1.pos, m.pos))
        self.assertAlmostEqual(ns.dist(a1.pos, m1.pos), 0, delta=5e-6)
        self.assertAlmostEqual(ns.dist(a1.pos, m2.pos), 0.13, delta=5e-3)
        cra2 = m2.to_cra(st[0])
        self.assertEqual(cra2.chain.name, 'B')
        self.assertEqual(str(cra2.residue.seqid), '37')
        self.assertEqual(cra2.atom.name, 'SG')
        marks2 = ns.find_neighbors(a1, 0.1, 3)
        self.assertEqual(len(marks2), 1)
        self.assertEqual(marks2[0], m2)

    def test_5a11_using_add_atom(self):
        self.test_5a11(use_populate=False)

    def test_1gtv(self):
        st = gemmi.read_pdb_string(FRAGMENT_1GTV)
        a1 = st[0].sole_residue('A', gemmi.SeqId(85, ' '))[0]
        ns = gemmi.NeighborSearch(st[0], st.cell, 5)
        ns.populate()
        marks = ns.find_atoms(a1.pos, a1.altloc, radius=3)
        self.assertEqual(len(marks), 2)
        for mark in marks:
            d = ns.dist(a1.pos, mark.pos)
            self.assertAlmostEqual(d, 0, delta=5e-6)
        marks2 = ns.find_neighbors(a1, 0.1, 3)
        self.assertEqual(len(marks2), 0)

        pos = gemmi.Position(100, 150, 200)
        mark = ns.find_nearest_atom(pos)
        cra = mark.to_cra(st[0])
        nim = st.cell.find_nearest_pbc_image(pos, mark.pos, 0)
        p = st.cell.find_nearest_pbc_position(cra.atom.pos, pos,
                                              mark.image_idx, inverse=True)
        self.assertAlmostEqual(nim.dist(), p.dist(cra.atom.pos))

    def test_b208(self):
        st = gemmi.read_structure(full_path('4oz7.pdb'))
        hoh208 = gemmi.Selection('B/208').copy_model_selection(st[0])
        self.assertEqual(hoh208.count_atom_sites(), 1)
        point = hoh208[0][0][0].pos
        for max_radius in [5, 20]:
            ns = gemmi.NeighborSearch(st[0], st.cell, max_radius).populate()
            marks = ns.find_atoms(point, min_dist=1e-6, radius=3)
            self.assertEqual(len(marks), 2)
            images = [m.image_idx for m in marks]
            self.assertEqual(sorted(images), [2, 3])
            image2 = st.cell.find_nearest_pbc_image(point, point, 2)
            self.assertEqual(image2.symmetry_code(), '3_545')
            self.assertAlmostEqual(image2.dist(), 2.9710496, delta=1e-6)
            image3 = st.cell.find_nearest_pbc_image(point, point, 3)
            self.assertEqual(image3.symmetry_code(), '4_355')


class TestContactSearch(unittest.TestCase):
    def test_radii_setting(self):
        cs = gemmi.ContactSearch(4.0)
        hg = gemmi.Element('Hg')
        self.assertEqual(cs.get_radius(hg), 0)
        cs.setup_atomic_radii(1, 0)
        cs.set_radius(hg, 1.5)
        self.assertEqual(cs.get_radius(hg), 1.5)

    def test_ignore_flag(self):
        st = gemmi.read_structure(full_path('4oz7.pdb'))
        st.setup_entities()
        ns = gemmi.NeighborSearch(st[0], st.cell, 5).populate()
        cs = gemmi.ContactSearch(4.0)
        cs.ignore = gemmi.ContactSearch.Ignore.SameResidue
        results = cs.find_contacts(ns)
        self.assertEqual(len(results), 607)
        cs.ignore = gemmi.ContactSearch.Ignore.SameChain
        results = cs.find_contacts(ns)
        self.assertEqual(len(results), 190)
        self.assertTrue(all(r.image_idx != 0
                            or r.partner1.chain is not r.partner2.chain
                            for r in results))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import os
import unittest
import gemmi

def full_path(filename):
    return os.path.join(os.path.dirname(__file__), filename)

# tests for gemmi/smcif.hpp
class TestRealCif(unittest.TestCase):
    def test_cod_sic(self):
        path = full_path('1011031.cif')
        sic = gemmi.read_small_structure(path)
        self.assertEqual(sic.name, '1011031')
        self.assertEqual(sic.cell.a, 4.358)
        self.assertEqual(sic.cell.alpha, 90.0)
        self.assertEqual(len(sic.sites), 2)
        self.assertEqual(len(sic.get_all_unit_cell_sites()), 8)
        self.assertEqual(4 * sic.sites[1].orth(sic.cell).x, sic.cell.a)

    def test_fen4(self):
        path = full_path('2242624.cif')
        small = gemmi.read_small_structure(path)
        types = small.atom_types
        self.assertEqual(len(types), 2)
        self.assertEqual(types[0].symbol, 'Fe')
        self.assertEqual(types[1].element, gemmi.Element('N'))

    def test_perovskite(self):
        small = gemmi.read_small_structure(full_path('4003024.cif'))
        self.assertEqual(small.spacegroup_hm, 'P m -3 m')
        disorder_groups = [site.disorder_group for site in small.sites]
        self.assertEqual(disorder_groups, [0, 1, 0, 2])

    def test_mgi2(self):
        small = gemmi.read_small_structure(full_path('2013551.cif'))
        for site in small.sites:
            u_eq = small.cell.calculate_u_eq(site.aniso)
            self.assertAlmostEqual(u_eq, site.u_iso, delta=0.00012)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
import os
import sys
import argparse
from collections import Counter
from gemmi import cif


def to_formula(cnt):
    '{C:12, O:8, N:1} -> "C12 N O8"'
    return ' '.join(k + (str(v) if v != 1 else '')
                    for k, v in sorted(cnt.items()))

def formula_to_dict(formula):
    '"O4 P -3" -> {O:4, P:1}'
    fdict = {}
    for elnum in formula.split():
        na = sum(e.isalpha() for e in elnum)
        if na == len(elnum):
            fdict[elnum] = 1
        elif na != 0:
            fdict[elnum[:na]] = int(elnum[na:])
    return fdict

def get_monomer_cifs(mon_path):
    'yield all files mon_path/?/*.cif in alphabetic order'
    for root, dirs, files in os.walk(mon_path):
        dirs.sort()
        for name in sorted(files):
            # in $CLIBD_MON, files with '_' in the name are not monomers
            if name.endswith('.cif') and '_' not in name:
                yield os.path.join(root, name)
            elif name.endswith('.cif.gz'):
                yield os.path.join(root, name)

def check_formulas(ccd):
    '''\
    Print the chemical components in CCD (components.cif) that have
    formula not consistent with the list of atoms.
    '''
    for b in ccd:
        atoms = Counter(a.upper()
                        for a in b.find_values('_chem_comp_atom.type_symbol'))
        formula = cif.as_string(b.find_value('_chem_comp.formula')).upper()
        fdict = formula_to_dict(formula)
        if fdict != atoms:
            print('[%s]' % b.name, formula, '<>', to_formula(atoms))


def compare_monlib_with_ccd(mon_path, ccd, verbose=False):
    'compare monomers from monomer library and CCD that have the same names'
    PRINT_MISSING_ENTRIES = False
    cnt = 0
    for path in get_monomer_cifs(mon_path):
        mon = cif.read(path)
        for mb in mon:
            if mb.name in ('', 'comp_list'):
                continue
            assert mb.name.startswith('comp_')
            name = mb.name[5:]
            cb = ccd.find_block(name)
            if cb:
                if verbose:
                    print('Comparing', cb.name)
                compare_chem_comp(mb, cb)
                cnt += 1
            elif PRINT_MISSING_ENTRIES:
                print('Not in CCD:', name)
    print('Compared', cnt, 'monomers.')


def get_heavy_atom_names(block):
    cca = block.find('_chem_comp_atom.', ['atom_id', 'type_symbol'])
    d = {a.str(0).upper(): a.str(1).upper() for a in cca if a[1] != 'H'}
    assert len(d) == sum(a[1] != 'H' for a in cca), (d, cca)
    return d


def bond_info(id1, id2, order, aromatic):
    if id1 > id2:
        id1, id2 = id2, id1
    order = order[:4].lower()
    assert order in ['sing', 'doub', 'trip', 'arom', 'delo', '1.5'], order
    aromatic = aromatic.upper()
    assert aromatic in ('Y', 'N'), aromatic
    if order in ('sing', 'doub') and aromatic == 'Y':
        order = 'arom'
    return ((id1, id2), (order, aromatic))


def bond_dict(block, ccb_names, atom_names):
    table = block.find('_chem_comp_bond.', ccb_names)
    return dict(bond_info(r.str(0), r.str(1), r.str(2), r.str(3))
                for r in table
                if r.str(0) in atom_names and r.str(1) in atom_names)


def compare_chem_comp(mb, cb):
    mon_names = get_heavy_atom_names(mb)
    ccd_names = get_heavy_atom_names(cb)

    # compare atom names (and counts)
    if mon_names != ccd_names:
        print('atom names differ in', cb.name)
        mon_atoms = Counter(mon_names.values())
        ccd_atoms = Counter(ccd_names.values())
        if mon_atoms != ccd_atoms:
            # can be relaxed by: and mon_atoms + Counter('O') != ccd_atoms
            print(cb.name, to_formula(mon_atoms), '<>', to_formula(ccd_atoms))

    # compare bonds
    mbonds = bond_dict(mb, ['atom_id_1', 'atom_id_2', 'type', 'aromatic'],
                       mon_names)
    cbonds = bond_dict(cb, ['atom_id_1', 'atom_id_2', 'value_order',
                            'pdbx_aromatic_flag'],
                       ccd_names)
    if mbonds != cbonds:
        print('Bonds differ in', cb.name)
        if set(mbonds) != set(cbonds):
            print('  extra bonds in mon.lib.:',
                  ' '.join('%s-%s' % p for p in set(mbonds) - set(cbonds)))
            print('  extra bonds in CCD:',
                  ' '.join('%s-%s' % p for p in set(cbonds) - set(mbonds)))
        for apair in sorted(set(mbonds) & set(cbonds)):
            if mbonds[apair] != cbonds[apair]:
                print('  %s-%s: %s/%s <> CCD %s/%s' %
                      (apair + mbonds[apair] + cbonds[apair]))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('ccd_path', metavar='/path/to/components.cif[.gz]')
    parser.add_argument('-m', metavar='DIR',
                        help='monomer library path (default: $CLIBD_MON)')
    parser.add_argument('-f', action='store_true', help='check CCD formulas')
    parser.add_argument('-v', action='store_true', help='verbose')
    args = parser.parse_args()
    mon_path = args.m or os.getenv('CLIBD_MON')
    if not mon_path and not args.f:
        sys.exit('Unknown monomer library path: use -m or set $CLIBD_MON.')
    ccd = cif.read(args.ccd_path)
    if args.f:
        check_formulas(ccd)
    if mon_path:
        compare_monlib_with_ccd(mon_path, ccd, verbose=args.v)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

import gzip
from io import StringIO
import os
import pickle
import unittest

import gemmi
from common import full_path, get_path_for_tempfile, numpy
try:
    from Bio import PDB
except ImportError:
    PDB = None

def is_written_to_pdb(line, via_cif):
    if line[:6] in ['COMPND', 'SOURCE', 'MDLTYP', 'AUTHOR', 'REVDAT', 'JRNL  ',
                    'SEQADV', 'HETNAM', 'FORMUL',
                    'SITE  ', 'MASTER', 'CONECT']:
        return False
    # ORIGX is written only if it is a non-identity matrix
    # SCALE is written only if it is non-default
    if line[:5] in ['ORIGX', 'SCALE']:
        return False
    if line[:6] == 'REMARK' and via_cif and line[7:10] not in ['  2', '350']:
        return False
    return True

# $ zgrep -P '(^HEADER|CRYST1|SSBOND|ATOM.*SG)' pdb5cfg.ent.gz
SSBOND_FRAGMENT = """\
HEADER    LYASE                                   08-JUL-15   5CFG              
SSBOND   1 CYS A  138    CYS A  138                          1555   2555  2.48  
CRYST1   86.530   45.120   77.980  90.00 105.15  90.00 C 1 2 1       4          
ATOM    166  SG  CYS A  65      17.771 -16.223 -10.059  1.00 39.15           S  
ATOM    396  SG  CYS A  93      10.163 -20.624 -11.577  1.00 25.81           S  
ATOM    444  SG  CYS A  99      12.757 -34.484  -2.237  1.00 44.66           S  
ATOM    744  SG  CYS A 138       0.524 -16.872  -1.123  1.00 38.43           S  
ATOM   1309  SG  CYS A 208      10.668 -20.500 -14.891  1.00 25.60           S  
ATOM   2024  SG  CYS A 296      14.668  -4.407 -16.359  1.00 25.70           S  
ATOM   2128  SG  CYS A 310      24.141 -22.158 -17.213  1.00 20.07           S  
"""  # noqa: W291 - trailing whitespace

SHORT_SSBOND = """\
SSBOND   1 CYS A    6    CYS A   11
ATOM     37  SG  CYS A   6      38.416  25.985  18.085  1.00 14.07  16       S
ATOM     69  SG  CYS A  11      36.989  25.994  19.570  1.00 14.23  16       S
"""

# fragment of $CCP4/examples/toxd/toxd_mod_p1.pdb without chain names
BLANK_CHAIN_FRAGMENT = """\
CRYST1   50.000   50.00    50.00   90.00  90.00   90.00
ATOM      1  N   GLN     1      16.207  -7.425   8.244  1.00 35.23
ATOM      2  CA  GLN     1      16.175  -6.019   8.805  1.00 35.13
ATOM      3  C   GLN     1      14.890  -5.752   9.585  1.00 33.49
ATOM      4  O   GLN     1      14.563  -6.517  10.541  1.00 35.33
ATOM      5  N   PRO     2      14.235  -4.626   9.270  1.00 30.05
ATOM      6  CA  PRO     2      12.967  -4.255   9.939  1.00 26.63
ATOM      7  C   PRO     2      11.955  -5.328   9.522  1.00 24.15
ATOM      8  O   PRO     2      12.217  -6.028   8.540  1.00 22.83
ATOM      9  CB  PRO     2      12.617  -2.895   9.427  1.00 26.18
ATOM     10  CG  PRO     2      13.342  -2.761   8.127  1.00 28.07
ATOM     11  CD  PRO     2      14.574  -3.663   8.212  1.00 29.12
ATOM     12  N   ARG     3      10.892  -5.446  10.311  1.00 21.86
ATOM     13  CA  ARG     3       9.875  -6.426  10.119  1.00 19.82
"""

# from https://github.com/project-gemmi/gemmi/issues/37
AMBER_FRAGMENT = """\
ATOM      7  CB  VAL     1     -14.375 -11.856  27.866  1.00  0.00
ATOM      8  HB  VAL     1     -14.217 -11.118  27.080  1.00  0.00
ATOM      9  CG1 VAL     1     -13.033 -12.232  28.471  1.00  0.00
ATOM     10 HG11 VAL     1     -12.398 -12.673  27.702  1.00  0.00
"""

FRAGMENT_WITH_HG = """\
HETATM 4406 HG    HG P 693      28.820  31.751  40.919  0.20 25.99
HETATM 4407 HG1   HG P 694      27.455  32.086  39.686  0.20 35.18
"""

TRJCONV_FRAGMENT = """\
ATOM  12609 5C'N NPH  6378     421.300 400.390 491.570  1.00  0.00
ATOM  12610 1HN5 NPH  6378     422.020 400.650 490.790  1.00  0.00
ATOM  12611 2HN5 NPH  6378     420.490 399.940 490.980  1.00  0.00
ATOM  12612 5O'N NPH  6378     421.940 399.470 492.410  1.00  0.00
ATOM  12613  PN  NPH  6378     422.280 397.980 491.900  1.00  0.00
ATOM  12614 1OPN NPH  6378     422.690 397.240 493.120  1.00  0.00
ATOM  12615 2OPN NPH  6378     423.260 398.170 490.810  1.00  0.00
ATOM  12616  O3P NPH  6378     420.850 397.490 491.350  1.00  0.00
"""


# from $CCP4/examples/data/insulin.pdb
UNORDERED_ALTLOC_FRAGMENT = """\
ATOM     54  CB  THR A   8      21.486  49.557  34.680  1.00 17.33           C  
ANISOU   54  CB  THR A   8     2677   1737   2168   -147    348    -77       C  
ATOM     55  N  ASER A   9      22.340  46.718  36.111  0.50 15.13           N  
ANISOU   55  N  ASER A   9     2397   1624   1726    -74    278    -66       N  
ATOM     56  CA ASER A   9      22.426  45.878  37.287  0.50 14.83           C  
ANISOU   56  CA ASER A   9     2219   1729   1687    -83    202    -92       C  
ATOM     57  C  ASER A   9      23.045  44.544  36.849  0.50 13.99           C  
ANISOU   57  C  ASER A   9     2072   1701   1540   -107    221    -92       C  
ATOM     58  O  ASER A   9      23.264  44.329  35.667  0.50 13.72           O  
ANISOU   58  O  ASER A   9     2090   1725   1397   -111    331    -63       O  
ATOM     59  CB ASER A   9      21.034  45.674  37.848  0.50 15.28           C  
ANISOU   59  CB ASER A   9     2236   1828   1740      0    251   -242       C  
ATOM     60  OG ASER A   9      20.347  46.901  38.012  0.50 19.98           O  
ANISOU   60  OG ASER A   9     2799   2271   2521    150    548     90       O  
ATOM     61  N  BSER A   9      22.313  46.728  36.138  0.50 14.71           N  
ANISOU   61  N  BSER A   9     2322   1562   1703    -78    282    -86       N  
ATOM     62  CA BSER A   9      22.511  45.877  37.303  0.50 14.08           C  
ANISOU   62  CA BSER A   9     2078   1645   1625   -106    203   -118       C  
ATOM     63  C  BSER A   9      22.951  44.499  36.837  0.50 13.55           C  
ANISOU   63  C  BSER A   9     1971   1640   1538   -117    232   -119       C  
ATOM     64  O  BSER A   9      22.899  44.194  35.639  0.50 12.89           O  
ANISOU   64  O  BSER A   9     1807   1651   1438    -89    415    -90       O  
ATOM     65  CB BSER A   9      21.244  45.783  38.133  0.50 14.46           C  
ANISOU   65  CB BSER A   9     1987   1655   1851    -61    227   -231       C  
ATOM     66  OG BSER A   9      20.199  45.110  37.441  0.50 13.08           O  
ANISOU   66  OG BSER A   9     1812   1758   1398    -16    450     56       O  
ATOM     67  N   VAL A  10      23.342  43.662  37.798  1.00 13.44           N  
ANISOU   67  N   VAL A  10     2079   1653   1371   -123    133   -148       N  
"""  # noqa: W291 - trailing whitespace

# from https://cci.lbl.gov/hybrid_36/
HY36_EXAMPLE = b"""\
ATOM  99998  SD  MET L9999      48.231 -64.383  -9.257  1.00 11.54           S
ATOM  99999  CE  MET L9999      49.398 -63.242 -10.211  1.00 14.60           C
ATOM  A0000  N   VAL LA000      52.228 -67.689 -12.196  1.00  8.76           N
ATOM  A0001  CA  VAL LA000      53.657 -67.774 -12.458  1.00  3.40           C
"""

# from 1keb.pdb
TER_EXAMPLE = """\
ATOM   1643  OXT ALA B 108      -2.885   7.940  32.034  1.00 44.47           O  
TER    1644      ALA B 108                                                      
HETATM 1645 CU    CU A 109       3.422  -7.286  12.794  1.00 20.53          CU  
HETATM 1646 CU    CU B 109       3.446  18.122   8.201  1.00 15.05          CU  
"""  # noqa: W291 - trailing whitespace

def read_lines_and_remove(path):
    with open(path) as f:
        out_lines = f.readlines()
    os.remove(path)
    return out_lines


class TestMol(unittest.TestCase):
    def test_atom(self):
        atom = gemmi.Atom()
        atom.occ = 0.5
        atom.element = gemmi.Element('C')
        atom.name = "CA"
        atom.pos = (-3, 4.5, 0)
        self.assertEqual(atom.occ, 0.5)
        self.assertEqual(atom.element.name, 'C')
        self.assertEqual(atom.name, 'CA')
        self.assertEqual(atom.pos.tolist(), [-3, 4.5, 0])
        atom.pos = gemmi.Position(0, -2, 2.5)
        self.assertEqual(atom.pos.tolist(), [0, -2, 2.5])

    def test_residue(self):
        res = gemmi.Residue()
        self.assertEqual(res.label_seq, None)
        self.assertEqual(res.seqid.num, None)
        res.label_seq = 1
        self.assertEqual(res.label_seq, 1)
        res.seqid.num = 2
        self.assertEqual(res.seqid.num, 2)
        res.label_seq = None
        self.assertEqual(res.label_seq, None)
        res.seqid.num = None
        self.assertEqual(res.seqid.num, None)
        for name in ['HOH', 'hoh', 'DOD', 'h2o', 'H2O', 'WAT']:
            res.name = name
            self.assertTrue(res.is_water())
        for name in ['SO4', '', 'HO', 'hoho', 'oho']:
            res.name = name
            self.assertFalse(res.is_water())

    def test_read_5i55(self):
        cell = gemmi.read_structure(full_path('5i55.cif')).cell
        self.assertAlmostEqual(cell.a, 29.46)
        self.assertAlmostEqual(cell.b, 10.51)
        self.assertAlmostEqual(cell.c, 29.71)
        self.assertEqual(cell.alpha, 90)
        self.assertAlmostEqual(cell.beta, 111.98)
        self.assertEqual(cell.gamma, 90)

    def test_read_5i55_again(self):
        block = gemmi.cif.read(full_path('5i55.cif'))[0]
        st = gemmi.make_structure_from_block(block)
        self.assertEqual(st.info['_entry.id'], '5I55')

        center = st[0].calculate_center_of_mass()
        # PyMOL>print cmd.centerofmass()
        pymol_ctr = [15.468438991742687, 4.8312495347721045, 20.607400844016833]
        self.assertTrue(center.dist(gemmi.Position(*pymol_ctr)) < 1e-7)

        chain, = st[0]
        a, b, c, d = chain.subchains()
        ent_a = st.get_entity_of(a)
        self.assertEqual(ent_a.name, '1')
        self.assertEqual(ent_a.entity_type, gemmi.EntityType.Polymer)
        self.assertEqual(ent_a.polymer_type, gemmi.PolymerType.PeptideL)
        ent_b = st.get_entity_of(b)
        self.assertEqual(ent_b.entity_type, gemmi.EntityType.NonPolymer)
        self.assertEqual(ent_b.polymer_type, gemmi.PolymerType.Unknown)
        ent_d = st.get_entity('4')
        self.assertEqual(ent_d.subchains, ['D'])
        self.assertEqual(ent_d.entity_type, gemmi.EntityType.Water)
        self.assertEqual(ent_d.polymer_type, gemmi.PolymerType.Unknown)

        output_block = st.make_mmcif_document().sole_block()
        cnames = block.get_mmcif_category_names()
        cnames_out = [name for name in output_block.get_mmcif_category_names()
                      if len(output_block.find_mmcif_category(name)) > 0]
        common_categories = [name for name in cnames_out if name in cnames]
        common_categories.sort()
        cc = ['_atom_site.', '_atom_type.', '_audit_author.',
              '_cell.', '_chem_comp.',
              '_diffrn.', '_diffrn_detector.', '_diffrn_radiation.',
              '_diffrn_source.',
              '_entity.', '_entity_poly.', '_entity_poly_seq.', '_entry.',
              '_exptl.', '_exptl_crystal.', '_pdbx_database_status.',
              '_pdbx_struct_assembly.', '_pdbx_struct_assembly_gen.',
              '_pdbx_struct_mod_residue.',
              '_pdbx_struct_oper_list.', '_refine.', '_reflns.', '_software.',
              '_struct.', '_struct_asym.', '_struct_conf.',
              '_struct_conf_type.', '_struct_conn.', '_struct_conn_type.',
              '_struct_keywords.', '_struct_ref.', '_struct_ref_seq.',
              '_symmetry.']
        self.assertEqual(common_categories, cc)
        mismatching = {'_struct_conf.id', '_chem_comp.type',
                       '_refine.ls_number_reflns_R_work'}
        for name in common_categories:
            cat_in = block.get_mmcif_category(name)
            cat_out = output_block.get_mmcif_category(name)
            for tag, values_out in cat_out.items():
                if tag == 'ccp4_link_id':
                    continue
                values_in = cat_in[tag]
                self.assertEqual(len(values_in), len(values_out))
                for (a, b) in zip(values_in, values_out):
                    try:
                        if a == b or abs(float(a) - float(b)) < 2e-4:
                            continue
                    except (ValueError, TypeError):
                        pass
                    self.assertTrue(name+tag in mismatching, name+tag)
        for name_out in cnames_out:
            self.assertTrue(name_out in cnames)

    def test_5i55_predefined_removals(self, clear_entities=False):
        st = gemmi.read_structure(full_path('5i55.cif'))
        if clear_entities:
            self.assertEqual(len(st.entities), 4)
            st.entities = gemmi.EntityList()
            self.assertEqual(len(st.entities), 0)
        lys12 = st[0]['A']['12']['LYS']
        count_b = sum(a.altloc == 'B' for a in lys12)
        model = st[0]
        # one author-chain and 4 label-chains: AA, 2 x ligand, waters
        self.assertEqual(len(model), 1)
        self.assertEqual(len(model.subchains()), 4)
        n_waters = len(model.get_subchain('D'))
        n_sites = model.count_atom_sites()
        occ_sum = model.count_occupancies()
        self.assertEqual(n_sites, occ_sum + count_b)
        st.remove_waters()
        self.assertEqual(len(model.subchains()), 3)
        self.assertEqual(model.count_atom_sites(), n_sites - n_waters)
        self.assertEqual(model.count_occupancies(), occ_sum - n_waters)
        st.remove_empty_chains()
        self.assertEqual(len(model.subchains()), 3)
        n_res = len(model['A'])
        st.remove_ligands_and_waters()
        self.assertEqual(len(model['A']), n_res - 2)
        mse = model['A'][0]
        self.assertEqual(len(mse), 8)
        mse.trim_to_alanine()
        self.assertEqual([a.name for a in mse], ['N', 'CA', 'C', 'O', 'CB'])
        model['A'].trim_to_alanine()
        # ALA has 5 atoms, except the last one which has OXT (hence +1)
        expected_count = sum(4 + (r.name != 'GLY') for r in model['A']) + 1
        self.assertEqual(model.count_occupancies(), expected_count)
        st.remove_alternative_conformations()
        self.assertEqual(model.count_occupancies(), expected_count - 5/2.)
        self.assertTrue(not any(a.has_altloc() for a in lys12))

    def test_5i55_predefined_removals2(self):
        self.test_5i55_predefined_removals(clear_entities=True)

    def test_rnase_predefined_removals(self, add_entities=False):
        st = gemmi.read_structure(full_path('rnase_frag.pdb'))
        if add_entities:
            self.assertEqual(len(st.entities), 0)
            st.add_entity_types()
            st.assign_subchains()
            st.ensure_entities()
            self.assertEqual(len(st.entities), 4)
        model = st[0]
        nres_a = len(model['A'])
        nres_b = len(model['B'])
        st.add_entity_types()
        st.remove_ligands_and_waters()  # removes SO4 from each chain
        self.assertEqual(len(model['A']), nres_a - 1)
        self.assertEqual(len(model['B']), nres_b - 1)
        self.assertEqual(len(model['W']), 0)
        self.assertEqual(len(model), 3)
        st.remove_empty_chains()
        self.assertEqual(len(model), 2)

    def test_rnase_predefined_removals2(self):
        self.test_rnase_predefined_removals(add_entities=True)

    def test_3dg1(self):
        st = gemmi.read_structure(full_path('3dg1_final.cif'))
        self.assertEqual(st.info['_entry.id'], '3DG1')
        self.assertEqual(len(st[0]), 1)
        chain = st[0]['A']
        for res in chain[-2:]:
            self.assertEqual(res.name, 'HOH')
        for res in chain:
            for atom in res:
                n_images = st.cell.is_special_position(atom.pos)
                self.assertEqual(atom.occ * (n_images + 1), 1.0)

    def test_3wup(self):
        st = gemmi.read_structure(full_path('3wup.json.gz'))
        self.assertEqual(st.info['_entry.id'], '3WUP')
        model = st[0]
        self.assertEqual(len(model), 1)
        self.assertEqual(len(st.assemblies), 2)

        special_count = 0
        for res in st[0]['A'].get_ligands():
            for atom in res:
                n_images = st.cell.is_special_position(atom.pos)
                if n_images:
                    special_count += 1
                    self.assertEqual(round(atom.occ, 6), 0.33)
        self.assertEqual(special_count, 2)

        site_count = model.count_atom_sites()
        self.assertEqual(site_count, 279)
        how = gemmi.HowToNameCopiedChain.Short
        a1 = gemmi.make_assembly(st.assemblies[0], model, how)
        self.assertEqual(a1.count_atom_sites(), site_count)
        a2 = gemmi.make_assembly(st.assemblies[1], model, how)
        self.assertEqual(a2.count_atom_sites(), site_count * 3)
        gemmi.merge_atoms_in_expanded_model(a2, gemmi.UnitCell())
        # 3 atoms are on a 3-fold rotation axis
        self.assertEqual(a2.count_atom_sites(), (site_count - 3) * 3 + 3 * 1)

    def test_software_category(self):
        doc = gemmi.cif.read_file(full_path('3dg1_final.cif'))
        input_block = doc.sole_block()
        st = gemmi.make_structure_from_block(input_block)
        output_block = st.make_mmcif_document().sole_block()
        software = output_block.get_mmcif_category('_software')
        assert software['version'][0] is False
        software['version'][0] = None
        self.assertEqual(input_block.get_mmcif_category('_software'), software)

    def test_5moo_header(self):
        st = gemmi.read_structure(full_path('5moo_header.pdb'))
        block = st.make_mmcif_document().sole_block()
        refine = block.get_mmcif_category('_refine')
        self.assertEqual(refine['ls_d_res_high'], ['1.44', '1.43'])
        self.assertEqual(refine['pdbx_starting_model'], ['4I8H'] * 2)
        self.assertEqual(list(block.find_values('_diffrn.ambient_temp')),
                         ['295', '295'])

    def check_1pfe(self, st):
        self.assertAlmostEqual(st.cell.a, 39.374)
        self.assertEqual(st.cell.gamma, 120)
        self.assertEqual(st.name, '1PFE')
        self.assertEqual(st.spacegroup_hm, 'P 63 2 2')
        self.assertEqual(st.info['_entry.id'], '1PFE')
        self.assertEqual(len(st), 1)
        self.assertEqual(len(st[0]), 2)
        label_to_auth_name = {sub.subchain_id(): ch.name for ch in st[0]
                              for sub in ch.subchains()}
        self.assertEqual(label_to_auth_name,
                         dict(A='A', B='B', C='A', D='B', E='B', F='A', G='B'))
        self.assertEqual(len(st[0]['A']['1']), 1)
        chain_a = st[0]['A']
        self.assertEqual(chain_a[0].label_seq, 1)
        self.assertEqual(chain_a['1'][0].seqid.num, 1)
        b3 = st[0]['B']['3']
        self.assertEqual(len(b3), 2)
        self.assertEqual(repr(b3[0]), repr(st[0]['B'][2]))
        self.assertEqual(b3[0].name, 'N2C')
        self.assertEqual(b3[-1].name, 'NCY')
        chain_c = st[0].get_subchain('C')
        self.assertEqual(len(chain_c), 1)
        res_cl = list(chain_c)[0]
        self.assertEqual(res_cl.name, 'CL')
        self.assertEqual(len(res_cl), 1)
        atom_cl = res_cl.find_atom('CL', '*')
        self.assertAlmostEqual(atom_cl.occ, 0.17)
        self.assertEqual(atom_cl.element.name, 'Cl')
        return st

    def test_previous_next_residue(self):
        st = gemmi.read_structure(full_path('1pfe.cif.gz'),
                                  merge_chain_parts=False)
        chain_b = st[0]['B']
        res = chain_b['6']['ALA']
        res = chain_b.next_residue(res)
        self.assertEqual(res.name, 'NCY')
        res = chain_b.next_residue(res)
        self.assertEqual(res.name, 'MVA')
        self.assertIsNone(chain_b.next_residue(res))
        res = chain_b.previous_residue(res)
        self.assertEqual(res.name, 'NCY')
        res = chain_b.previous_residue(res)
        self.assertEqual(res.name, 'ALA')
        res = chain_b.next_residue(chain_b['7']['N2C'])
        self.assertEqual(res.name, 'MVA')
        self.assertEqual(chain_b.previous_residue(res).name, 'NCY')
        self.assertIsNone(chain_b.previous_residue(chain_b[0]))

    def test_read_1pfe_cif(self):
        st = gemmi.read_structure(full_path('1pfe.cif.gz'))
        self.check_1pfe(st)
        mmcif_doc = st.make_mmcif_document()

        # write structure to cif file and read it back
        out_name = get_path_for_tempfile(suffix='.cif')
        mmcif_doc.write_file(out_name)
        st2 = gemmi.read_structure(out_name)
        os.remove(out_name)
        self.check_1pfe(st2)

        # write structure to mmJSON string and read it back
        json_str = mmcif_doc.as_json(mmjson=True)
        doc = gemmi.cif.read_mmjson_string(json_str)
        st3 = gemmi.make_structure_from_block(doc[0])
        self.check_1pfe(st3)

    def test_read_1pfe_json(self):
        st = gemmi.read_structure(full_path('1pfe.json'))
        self.check_1pfe(st)

    def test_read_1orc(self):
        st = gemmi.read_structure(full_path('1orc.pdb'))
        self.assertEqual(st.resolution, 1.54)
        self.assertAlmostEqual(st.cell.a, 34.77)
        self.assertEqual(st.cell.alpha, 90)
        self.assertEqual(len(st.ncs), 0)
        self.assertEqual(st.meta.authors,
                         ['ALBRIGHT, R.A.', 'MOSSING, M.C.', 'MATTHEWS, B.W.'])
        model = st[0]
        self.assertEqual(len(model), 1)
        self.assertEqual(len(model.subchains()), 2)
        A = model['A']
        waters = A.get_waters()
        self.assertEqual(len(waters), 57)  # FORMUL   2  HOH   *57(H2 O)
        self.assertTrue(all(res.name == 'HOH' for res in waters))
        self.assertTrue(A['3'])
        self.assertFalse(A['0'])
        self.assertEqual([res.seqid.num for res in A if res.seqid.icode != ' '],
                         [56] * 5)
        self.assertEqual(len(A['55']), 1)
        self.assertEqual(len(A['55B']), 0)
        self.assertEqual(len(A['56B']), 1)
        self.assertEqual(A['56'][0].seqid.icode, ' ')
        self.assertEqual(A['56c'][0].seqid.icode, 'C')

        result = gemmi.align_sequence_to_polymer(st.entities[0].full_sequence,
                                                 A.get_polymer(),
                                                 gemmi.PolymerType.Unknown)
        self.assertEqual(result.cigar_str(), '2I64M5I')

    def write_and_read(self, st, via_cif):
        if via_cif:
            st.setup_entities()
            st.assign_label_seq_id()
            doc = st.make_mmcif_document()
            st = gemmi.make_structure_from_block(doc[0])
        out_name = get_path_for_tempfile()
        st.write_pdb(out_name)
        return read_lines_and_remove(out_name)

    def test_read_write_1orc(self, via_cif=False):
        path = full_path('1orc.pdb')
        with open(path) as f:
            expected = [line for line in f if is_written_to_pdb(line, via_cif)]
        st = gemmi.read_structure(path)
        out_lines = self.write_and_read(st, via_cif)
        self.assertEqual(expected, out_lines)

    def test_read_write_1orc_via_cif(self):
        self.test_read_write_1orc(via_cif=True)

    def test_read_write_1lzh(self, via_cif=False):
        path = full_path('1lzh.pdb.gz')
        with gzip.open(path, mode='rt') as f:
            expected = [line for line in f if is_written_to_pdb(line, via_cif)]
        out_lines = self.write_and_read(gemmi.read_structure(path), via_cif)
        self.assertEqual(expected[0], out_lines[0])
        # TITLE lines differ because the text is broken at different word
        self.assertEqual(expected[3:], out_lines[3:])

    def test_read_write_1lzh_via_cif(self):
        self.test_read_write_1lzh(via_cif=True)

    def test_ncs_in_1lzh(self):
        st = gemmi.read_structure(full_path('1lzh.pdb.gz'))
        self.assertEqual(len(st.ncs), 1)
        A, B = st[0]
        for ra, rb in zip(A, B):
            pa = ra['CA'][0].pos
            pb = rb['CA'][0].pos
            image_of_pb = st.ncs[0].apply(pb)
            self.assertTrue(pa.dist(image_of_pb) < 0.01)

    def test_read_write_5cvz_final(self, via_cif=False):
        path = full_path('5cvz_final.pdb')
        with open(path) as f:
            expected = [line.rstrip() for line in f
                        if is_written_to_pdb(line, via_cif)
                        # SCALE is not written b/c CRYST1 has more precision.
                        and line[:5] != 'SCALE']
        st = gemmi.read_structure(path)
        if via_cif:
            # input file w/o TER record -> subchains not setup automatically
            st.setup_entities()
            doc = st.make_mmcif_document()
            st = gemmi.make_structure_from_block(doc[0])
        out_lines = self.write_and_read(st, via_cif=False)
        if via_cif:
            out_lines = [line for line in out_lines
                         # input file has no REMARK 2, but it gets generated
                         # from REMARK 3 when going pdb->cif->pdb
                         if line[:10] != 'REMARK   2'
                         and line[:5] != 'TER  ']
        self.assertEqual(expected, [line.rstrip() for line in out_lines])

    def test_read_write_5cvz_final_via_cif(self):
        self.test_read_write_5cvz_final(via_cif=True)

    def test_read_write_4oz7(self, via_cif=False):
        path = full_path('4oz7.pdb')
        with open(path) as f:
            expected = [line for line in f if is_written_to_pdb(line, via_cif)]
        st = gemmi.read_structure(path, merge_chain_parts=False)
        out_lines = self.write_and_read(st, via_cif)
        self.assertEqual(expected, out_lines)

    def test_read_write_4oz7_via_cif(self):
        self.test_read_write_4oz7(via_cif=True)

    @unittest.skipIf(PDB is None, "BioPython not installed.")
    def test_reading_output_mmcif_with_biopython(self):
        path = full_path('4oz7.pdb')
        st = gemmi.read_structure(path)
        groups = gemmi.MmcifOutputGroups(True)
        # BioPython parser chokes without _atom_site.group_PDB
        # groups.group_pdb = False
        doc = st.make_mmcif_document(groups)
        parser = PDB.MMCIFParser(QUIET=True)
        parser.get_structure("none", StringIO(doc.as_string()))

    def test_pdb_element_names(self):
        pdb_line = "HETATM 4154 MG    MG A 341       1.384  19.340  11.968" \
                   "  1.00 67.64          MG"
        for line in [pdb_line, pdb_line.strip(' MG'), pdb_line[:-2] + '  ']:
            st = gemmi.read_pdb_string(line)
            residue = st[0].sole_residue('A', gemmi.SeqId(341, ' '))
            mg_atom = residue.sole_atom('MG')
            self.assertEqual(mg_atom.element.name, 'Mg')
            self.assertEqual(mg_atom.padded_name(), 'MG')
            self.assertAlmostEqual(mg_atom.b_iso, 67.64, delta=1e-6)
        mg_atom.element = gemmi.Element('Cu')
        self.assertEqual(mg_atom.element.name, 'Cu')

    def test_pdb_misaligned_element(self):
        pdb_line = "ATOM      7 S    SUB A   7      34.489 -14.293  34.343" \
                   "  0.29 43.77          S"
        for line in [pdb_line, pdb_line + '\n', pdb_line + '\r\n']:
            st = gemmi.read_pdb_string(line)
            atom = st[0].sole_residue('A', gemmi.SeqId('7')).sole_atom('S')
            self.assertEqual(atom.element.name, 'S')

    def test_pdb_element_names_from_amber(self):
        st = gemmi.read_pdb_string(AMBER_FRAGMENT)
        residue = st[0][''][0]
        self.assertEqual(residue.sole_atom('CB').element, gemmi.Element('C'))
        self.assertEqual(residue.sole_atom('HB').element, gemmi.Element('H'))
        self.assertEqual(residue.sole_atom('CG1').element, gemmi.Element('C'))
        self.assertEqual(residue.sole_atom('HG11').element, gemmi.Element('H'))
        lines = AMBER_FRAGMENT.splitlines()
        for n, atom in enumerate(residue):
            self.assertEqual(atom.padded_name(), lines[n][12:16].rstrip())
        chain = gemmi.read_pdb_string(FRAGMENT_WITH_HG)[0]['P']
        self.assertEqual(chain[0].sole_atom('HG').element, gemmi.Element('Hg'))
        self.assertEqual(chain[1].sole_atom('HG1').element, gemmi.Element('Hg'))

    def test_pdb_element_names_from_trjconv(self):
        st = gemmi.read_pdb_string(TRJCONV_FRAGMENT)
        residue = st[0][''][0]
        expected = ['C', 'H', 'H', 'O', 'P', 'O', 'O', 'O']
        lines = TRJCONV_FRAGMENT.splitlines()
        for n, atom in enumerate(residue):
            self.assertEqual(atom.element.name, expected[n])
            self.assertEqual(atom.padded_name(), lines[n][12:16].rstrip())

    def test_4hhh_frag(self):
        path = full_path('4hhh_frag.pdb')
        with open(path) as f:
            frag = f.read()
        st = gemmi.read_pdb_string(frag)
        in_headers = frag.splitlines()
        out_headers = st.make_pdb_headers().splitlines()
        self.assertEqual(in_headers[0], out_headers[0])
        # the difference 4555 vs 2555 doesn't matter for us
        self.assertEqual(in_headers[1], out_headers[1].replace(' 4555 ',
                                                               ' 2555 '))
        self.assertEqual(in_headers[2], out_headers[2])

    def test_blank_mmcif(self):
        input_block = gemmi.cif.Block('empty')
        st = gemmi.make_structure_from_block(input_block)
        self.assertEqual(st.name, 'empty')
        output_block = st.make_mmcif_document().sole_block()
        self.assertEqual(output_block.get_mmcif_category_names(), [])

    def test_blank_chain(self):
        st = gemmi.read_pdb_string(BLANK_CHAIN_FRAGMENT)
        out_name = get_path_for_tempfile()
        st.write_minimal_pdb(out_name)
        out = read_lines_and_remove(out_name)
        # CRYST1 differs (50.000 not 50.00 and added P1).
        # ATOM lines have added element names.
        trimmed_out = [line[:66] for line in out[1:]]
        self.assertEqual(trimmed_out, BLANK_CHAIN_FRAGMENT.splitlines()[1:])

    def test_ncs(self):
        st = gemmi.read_structure(full_path('5cvz_final.pdb'))
        self.assertEqual(len(st.cell.images), 20 * 12 - 1)
        self.assertEqual(st.resolution, 3.29)
        first_atom = st[0].sole_residue('A', gemmi.SeqId(17, ' '))[0]
        ne2 = st[0].sole_residue('A', gemmi.SeqId('63')).sole_atom('NE2')
        direct_dist = first_atom.pos.dist(ne2.pos)
        self.assertAlmostEqual(direct_dist, 34.89, delta=1e-2)
        nearest_image = st.cell.find_nearest_image(first_atom.pos, ne2.pos)
        nearest_dist = nearest_image.dist()
        self.assertAlmostEqual(nearest_dist, 8.02, delta=1e-2)

        # test __getitem__(splice) - unrelated to NCS (sometimes we put
        # unrelated tests together to avoid reading the same file again)
        chain = st[0]['A']
        res = chain.next_residue(chain[:5][0])
        self.assertTrue(res is chain[1])

    def test_ssbond(self):
        st = gemmi.read_pdb_string(SSBOND_FRAGMENT)
        out = st.make_pdb_headers()
        self.assertEqual(out.splitlines(), SSBOND_FRAGMENT.splitlines()[:3])

    def test_ssbond_again(self):
        st = gemmi.read_pdb_string(SSBOND_FRAGMENT)
        doc = st.make_mmcif_document()
        st2 = gemmi.make_structure_from_block(doc[0])
        out = st2.make_pdb_headers()
        self.assertEqual(out.splitlines(), SSBOND_FRAGMENT.splitlines()[:3])

    def test_short_ssbond(self):
        st = gemmi.read_pdb_string(SHORT_SSBOND)
        out = st.make_pdb_headers()
        self.assertEqual(out.splitlines()[0],
                         SHORT_SSBOND.splitlines()[0]
                         + "                          1555   1555  2.06  ")

    def test_add_remove(self):
        st = gemmi.read_pdb_string(SSBOND_FRAGMENT)
        st.add_model(st[0])
        st.renumber_models()
        res = st[0].sole_residue('A', gemmi.SeqId('310'))
        self.assertEqual(len(res), 1)
        res.remove_atom('SG', ' ')
        self.assertEqual(len(res), 0)
        res = st[1].sole_residue('A', gemmi.SeqId('310'))
        self.assertEqual(len(res), 1)
        self.assertEqual(len(st[0]['A']), 7)
        del st[0]['A'][3]
        self.assertEqual(len(st[0]['A']), 6)
        self.assertEqual(len(st), 2)
        self.assertEqual(st[0].num, 1)
        del st[0]
        self.assertEqual(len(st), 1)
        self.assertEqual(st[0].num, 2)
        st.renumber_models()
        self.assertEqual(st[0].num, 1)
        st.add_model(st[0])
        st.add_model(st[0])
        st.renumber_models()
        self.assertEqual(st[0].num, 1)
        self.assertEqual(st[-1].num, 3)
        del st[:-1]
        self.assertEqual(len(st), 1)
        self.assertEqual(st[0].num, 3)
        del st[0]
        self.assertEqual(len(st), 0)

    def test_remove2(self):
        # test also save_doc
        saved_doc = gemmi.cif.Document()
        st = gemmi.read_structure(full_path('1pfe.cif.gz'), save_doc=saved_doc)
        self.assertEqual(saved_doc[0].name, '1PFE')
        self.assertEqual(saved_doc[0].find_value('_entity_name_com.name'),
                         "'QUINOMYCIN A'")
        model = st[0]
        self.assertEqual(len(model), 2)
        b = model['B']
        self.assertEqual(b[0].name, 'DSN')
        del b['1']['DSN']
        self.assertEqual(b[0].name, 'ALA')
        del b[0]
        self.assertEqual(b[0].name, 'N2C')

        # test append_residues()
        self.assertEqual(len(b), 20)
        b.append_residues(b[:5], min_sep=10)
        self.assertEqual(len(b), 25)

        # test append_residues() with empty chain
        new_chain = gemmi.Chain('X')
        new_chain.append_residues(b[:5], min_sep=1)
        self.assertEqual(len(new_chain), 5)

        # test adding and removing chains
        model.add_chain(new_chain, unique_name=False)
        model.add_chain(new_chain, unique_name=True)
        model.add_chain(new_chain)
        self.assertEqual([chain.name for chain in model], list('ABXCX'))
        del model[2:]
        model.add_chain(new_chain, unique_name=True)
        self.assertEqual([chain.name for chain in model], list('ABX'))
        del model[-1]
        del model['A']
        self.assertEqual(len(model), 1)
        self.assertEqual(model[0].name, 'B')
        doc = st.make_mmcif_document()
        ref_seq = doc[0].get_mmcif_category('_struct_ref_seq')
        self.assertEqual(ref_seq['pdbx_strand_id'], ['B'])

    def test_first_conformer(self):
        model = gemmi.read_structure(full_path('1pfe.cif.gz'))[0]
        b = model['B']
        self.assertEqual([res.name for res in b if not res.is_water()],
                         ['DSN', 'ALA', 'N2C', 'NCY', 'MVA', 'DSN',
                          'ALA', 'NCY', 'N2C', 'MVA', 'QUI', 'QUI'])
        self.assertEqual([res.name for res in b.first_conformer()
                          if not res.is_water()],
                         ['DSN', 'ALA', 'N2C', 'MVA', 'DSN',
                          'ALA', 'NCY', 'MVA', 'QUI', 'QUI'])
        polymer = b.get_polymer()
        self.assertEqual([res.name for res in polymer],
                         ['DSN', 'ALA', 'N2C', 'NCY', 'MVA', 'DSN',
                          'ALA', 'NCY', 'N2C', 'MVA'])
        self.assertEqual([res.name for res in polymer.first_conformer()],
                         ['DSN', 'ALA', 'N2C', 'MVA', 'DSN',
                          'ALA', 'NCY', 'MVA'])
        self.assertEqual(len(polymer), 10)
        self.assertEqual(polymer.length(), 8)
        # The bond between 4 MVA(v) and 5 DSN(s) is between C and OG
        # (so it's not a peptide bond as expected). Depending on the heuristic
        # used to determine gaps, this sequence could have a gap in the middle.
        self.assertEqual(polymer.make_one_letter_sequence(), 'sAXvsAXv')
        self.assertEqual([res.name for res in b.get_ligands()], ['QUI', 'QUI'])
        res1 = model.sole_residue('A', gemmi.SeqId('1'))
        self.assertEqual([atom.name for atom in res1.first_conformer()],
                         [atom.name for atom in res1 if atom.altloc != 'B'])

    def test_model_all(self):
        model = gemmi.Model(1)
        for name in 'ABCDEFG':
            model.add_chain(gemmi.Chain(name))
        expected = []
        for cname in 'BCF':
            chain = model[cname]
            for _ in range(7):
                chain.add_residue(gemmi.Residue())
            for (r, name) in [(2, '0'), (2, '1'), (3, '2'), (5, '3')]:
                a = gemmi.Atom()
                a.name = cname + name
                expected.append(a.name)
                chain[r].add_atom(a)
        self.assertEqual([cra.atom.name for cra in model.all()], expected)
        st = gemmi.Structure()
        st.add_model(model)
        st.remove_empty_chains()
        self.assertEqual([cra.atom.name for cra in model.all()], expected)

    def test_different_altloc_order(self):
        st = gemmi.read_pdb_string(UNORDERED_ALTLOC_FRAGMENT)
        chain = st[0]['A']
        cb = chain['9']['SER']['CB']
        cb_numbers = [atom.serial for atom in cb]
        self.assertEqual(cb_numbers, [59, 65])
        self.assertEqual(cb[0].serial, 59)
        self.assertEqual(cb[1].serial, 65)
        self.assertEqual(cb[-1].serial, 65)
        self.assertEqual(cb[-2].serial, 59)
        self.assertEqual(chain.count_atom_sites(), 14)
        self.assertEqual(chain.count_occupancies(), 8)
        st.remove_alternative_conformations()
        self.assertEqual(chain.count_atom_sites(), 8)

    def test_extract_sequence_info(self):
        st = gemmi.read_structure(full_path('5cvz_final.pdb'))
        st.add_entity_types()
        polymer = st[0][0].get_polymer()
        self.assertEqual(polymer.check_polymer_type(),
                         gemmi.PolymerType.PeptideL)
        expected = ('AAATSLVYDTCYVTLTERATTSFQRQSFPTLKGMGDRAFQVVAFTIQGVS'
                    'AAPLMYNARLYNPGDTDSVHATGVQLMGTVPRTVRLTPRVGQNNWFFGNT'
                    'EEAETILAIDGLVSTKGANAPSNTVIVTGCFRLAPSELQSS')
        self.assertEqual(polymer.make_one_letter_sequence(), expected)

    def test_assembly(self):
        st = gemmi.read_structure(full_path('1pfe.cif.gz'),
                                  merge_chain_parts=False)
        model = st[0]
        ch_names = [ch.name for ch in model]
        self.assertEqual(['A', 'B', 'A', 'B', 'A', 'B'], ch_names)
        model_mass = model.calculate_mass()

        self.assertEqual(len(st.assemblies), 1)
        asem = st.assemblies[0]
        bio = gemmi.make_assembly(asem, model,
                                  gemmi.HowToNameCopiedChain.Short)
        new_naming = {'A': 'C', 'B': 'D'}
        self.assertEqual([ch.name for ch in bio],
                         ch_names + [new_naming[x] for x in ch_names])
        self.assertAlmostEqual(bio.calculate_mass(), 2 * model_mass)

        bio = gemmi.make_assembly(asem, model,
                                  gemmi.HowToNameCopiedChain.AddNumber)
        self.assertEqual([ch.name for ch in bio],
                         [x+'1' for x in ch_names] + [x+'2' for x in ch_names])

    def test_assembly_naming(self):
        st = gemmi.read_structure(full_path('4oz7.pdb'))
        model = st[0]
        a1 = st.assemblies[1]
        bio = gemmi.make_assembly(a1, model,
                                  gemmi.HowToNameCopiedChain.AddNumber)
        self.assertEqual([ch.name for ch in bio], ['B1'])
        bio = gemmi.make_assembly(a1, model, gemmi.HowToNameCopiedChain.Dup)
        self.assertEqual([ch.name for ch in bio], ['B'])
        bio = gemmi.make_assembly(a1, model, gemmi.HowToNameCopiedChain.Short)
        self.assertEqual([ch.name for ch in bio], ['B'])

    def test_hybrid36(self):
        st = gemmi.read_pdb_string(HY36_EXAMPLE)
        nums = [(cra.atom.serial, cra.residue.seqid.num) for cra in st[0].all()]
        self.assertEqual(nums, [(99998, 9999), (99999, 9999),
                                (100000, 10000), (100001, 10000)])
        minimal_opt = gemmi.PdbWriteOptions(minimal=True)
        out = st.make_pdb_string(minimal_opt).splitlines()
        # original serial numbers are lost when writing pdb, check only seqid
        self.assertEqual(out[1][17:30], 'MET L9999    ')
        self.assertEqual(out[4][17:30], 'VAL LA000    ')

    def test_ter_writing(self):
        st = gemmi.read_pdb_string(TER_EXAMPLE)
        def write_and_read(**kwargs):
            opt = gemmi.PdbWriteOptions(**kwargs)
            kwargs['minimal'] = True
            st_out = gemmi.read_pdb_string(st.make_pdb_string(opt))
            return [(cra.chain.name + ' ' + cra.atom.name, cra.atom.serial)
                    for cra in st_out[0].all()]
        self.assertEqual(write_and_read(),
                         [('B OXT', 1), ('A CU', 3), ('B CU', 4)])
        st.merge_chain_parts()
        self.assertEqual(write_and_read(),
                         [('B OXT', 1), ('B CU', 3), ('A CU', 4)])
        self.assertEqual(write_and_read(numbered_ter=False),
                         [('B OXT', 1), ('B CU', 2), ('A CU', 3)])
        self.assertEqual(write_and_read(preserve_serial=True),
                         [('B OXT', 1643), ('B CU', 1646), ('A CU', 1645)])
        st.assign_serial_numbers(numbered_ter=True)
        self.assertEqual(write_and_read(preserve_serial=True),
                         [('B OXT', 1), ('B CU', 3), ('A CU', 4)])

    def test_serialization(self):
        st = gemmi.read_structure(full_path('5e5z.pdb'))
        unpickled = pickle.loads(pickle.dumps(st))
        self.assertEqual(st.make_pdb_string(), unpickled.make_pdb_string())
        unpickled[0] = pickle.loads(pickle.dumps(st[0]))
        self.assertEqual(st.make_pdb_string(), unpickled.make_pdb_string())
        chain = st[0][0]
        unpickled_chain = pickle.loads(pickle.dumps(chain))
        self.assertEqual(chain.name, unpickled_chain.name)
        self.assertEqual(len(chain), len(unpickled_chain))
        res = chain[0]
        unpickled_res = pickle.loads(pickle.dumps(res))
        self.assertEqual(res.name, unpickled_res.name)
        self.assertEqual(len(res), len(unpickled_res))
        atom = res[1]
        unpickled_atom = pickle.loads(pickle.dumps(atom))
        self.assertEqual(atom.name, unpickled_atom.name)

    def test_flat(self):
        st = gemmi.read_structure(full_path('5e5z.pdb'))
        fst = gemmi.FlatStructure(st)
        stback = fst.generate_structure()
        self.assertEqual(st.make_pdb_string(), stback.make_pdb_string())
        #self.assertEqual(pickle.dumps(st), pickle.dumps(stback))

    @unittest.skipIf(numpy is None, "requires NumPy")
    def test_flat_arrays(self):
        st = gemmi.read_structure(full_path('5e5z.pdb'))
        fst = gemmi.FlatStructure(st)

        # Test length
        n_atoms = len(fst)
        self.assertEqual(n_atoms, len(fst.b_iso))
        self.assertEqual(n_atoms, len(fst.occ))
        self.assertEqual(n_atoms, len(fst.atom_names))

        # Test shapes
        self.assertEqual(fst.pos.shape, (n_atoms, 3))
        self.assertEqual(fst.atom_names.shape, (n_atoms, 8))
        self.assertEqual(fst.residue_names.shape, (n_atoms, 8))
        self.assertEqual(fst.chain_ids.shape, (n_atoms, 8))
        fst.strings_as_numbers = False
        self.assertEqual(fst.atom_names.shape, (n_atoms,))

        # Test data types
        self.assertEqual(fst.b_iso.dtype, numpy.float32)
        self.assertEqual(fst.occ.dtype, numpy.float32)
        self.assertEqual(fst.pos.dtype, numpy.float64)
        # char arrays can be int8 or uint8 depending on platform
        #self.assertIn(fst.atom_names.dtype, [numpy.int8, numpy.uint8])

        # Test modifying arrays
        if n_atoms > 0:
            # Save original values
            orig_b_iso = fst.b_iso[0].copy()
            orig_occ = fst.occ[0].copy()
            orig_pos = fst.pos[0].copy()

            # Modify values
            fst.b_iso[0] = 99.9
            fst.occ[0] = 0.5
            fst.pos[0] = [1.0, 2.0, 3.0]

            # Verify changes
            self.assertAlmostEqual(fst.b_iso[0], 99.9, places=1)
            self.assertAlmostEqual(fst.occ[0], 0.5, places=1)
            numpy.testing.assert_array_almost_equal(fst.pos[0], [1.0, 2.0, 3.0])

            # Test that changes propagate to generated structure
            st_modified = fst.generate_structure()
            first_atom = next(cra.atom for cra in st_modified[0].all())
            self.assertAlmostEqual(first_atom.b_iso, 99.9, places=1)
            self.assertAlmostEqual(first_atom.occ, 0.5, places=1)
            numpy.testing.assert_array_almost_equal(first_atom.pos.tolist(),
                                                    [1.0, 2.0, 3.0])

            # Restore original values
            fst.b_iso[0] = orig_b_iso
            fst.occ[0] = orig_occ
            fst.pos[0] = orig_pos

    @unittest.skipIf(numpy is None, "requires NumPy")
    def test_flat_bulk_modification(self):
        st = gemmi.read_structure(full_path('5e5z.pdb'))
        fst = gemmi.FlatStructure(st)

        if len(fst) == 0:
            return  # Skip if no atoms

        # Set all B-factors to 20.0
        fst.b_iso[:] = 20.0

        # Verify all B-factors are now 20.0
        numpy.testing.assert_array_equal(fst.b_iso, 20.0)

        # Generate structure and verify B-factors propagated
        st_modified = fst.generate_structure()
        for cra in st_modified[0].all():
            self.assertAlmostEqual(cra.atom.b_iso, 20.0, places=1)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

# Read an archive of SF-mmCIF data and for each block print information:
# - PDB ID
# - 0-based block index
# - main category: R=_refln, D=_diffrn_refln
# - number of columns
# - number of rows
# - number of unique Miller indices (all symmetry-equivalent reflections
#   and Friedel pairs are counted as one)
# - do Miller indices repeat? N=no, F=only Friedel pairs, Y=yes

import argparse
import os
import gemmi

def process(path):
    doc = gemmi.cif.read(path)
    code = os.path.basename(path)[1:5]
    for n, rblock in enumerate(gemmi.as_refln_blocks(doc)):
        loop = rblock.default_loop
        assert loop is not None, 'missing tag index_h'
        cat = loop.tags[0][1].upper()
        (data_type, unique) = gemmi.check_data_type_under_symmetry(rblock)
        if data_type == gemmi.DataType.Mean:
            r = 'N'
        elif data_type == gemmi.DataType.Anomalous:
            r = 'F'
        elif data_type == gemmi.DataType.Unmerged:
            r = 'Y'
        else:  # gemmi.DataType.Unknown - when space group is missing
            r = 'X'
        width = loop.width()
        length = loop.length()
        print(f'{code}\t{n}\t{cat}\t{width}\t{length}\t{unique}\t{r}')

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--continue-from", type=str, help="Continue from file")
    args, remaining_args = parser.parse_known_args()
    for arg in remaining_args:
        for path in gemmi.CifWalk(arg, try_pdbid='S'):
            if args.continue_from is not None:
                if path != args.continue_from:
                    continue
                args.continue_from = None
            try:
                process(path)
            except Exception as e:
                print(path, 'was not processed:', e)
    if args.continue_from:
        print('Path not encountered:', args.continue_from)
        print('Last path:', path)

main()



#!/usr/bin/env python

import numpy
from matplotlib import pyplot
import gemmi

# toxd_aupatt.map is generated by $CCP4/examples/unix/runnable/patterson
ccp4 = gemmi.read_ccp4_map('/tmp/wojdyr/toxd_aupatt.map', setup=True)
arr = ccp4.grid.array
x = numpy.linspace(0, ccp4.grid.unit_cell.a, num=arr.shape[0], endpoint=False)
y = numpy.linspace(0, ccp4.grid.unit_cell.b, num=arr.shape[1], endpoint=False)
X, Y = numpy.meshgrid(x, y, indexing='ij')
pyplot.contour(X, Y, arr[:,:,40])
pyplot.gca().set_aspect('equal', adjustable='box')
pyplot.show()



# Compares serialization macros from the SERIALIZE_PATH file
# with libclang's AST of the TU_PATH translation unit.

from clang import cindex

SERIALIZE_PATH = 'include/gemmi/serialize.hpp'
TU_PATH = 'include/gemmi/model.hpp'

def read_macros_from_file():
    macros = {}
    with open(SERIALIZE_PATH) as f:
        name = None
        for line in f:
            line = line.strip()
            if line.startswith('SERIALIZE'):
                assert name is None
                name = line[line.index('(')+1 : line.index(',')]
                macros[name] = line
            elif name:
                macros[name] += ' ' + line
            if name and ')' in line:
                name = None
    return macros

def compare_struct(name, node, line_from_file):
    print(name, end=' ')
    struct_base = None
    template_params = []
    fields = []
    for child in node.get_children():
        if child.kind == cindex.CursorKind.TEMPLATE_TYPE_PARAMETER:
            template_params.append('typename')
        elif child.kind == cindex.CursorKind.TEMPLATE_NON_TYPE_PARAMETER:
            first_token = next(child.get_tokens())
            template_params.append(first_token.spelling)
        elif child.kind == cindex.CursorKind.CXX_BASE_SPECIFIER:
            (base_ref,) = tuple(child.get_children())
            assert base_ref.kind == cindex.CursorKind.TYPE_REF
            struct_base = base_ref.referenced.spelling
        elif child.kind == cindex.CursorKind.FIELD_DECL:
            # public = (child.access_specifier == cindex.AccessSpecifier.PUBLIC)
            fields.append(child.spelling)
    if node.kind == cindex.CursorKind.CLASS_TEMPLATE:
        assert len(template_params) == 1
        beginning = f'SERIALIZE_T1({name}, {template_params[0]}'
    else:
        if struct_base:
            beginning = f'SERIALIZE_P({name}, {struct_base}'
        else:
            beginning = f'SERIALIZE({name}'
    expected = beginning + ''.join(', o.' + field for field in fields) + ')'
    if expected == line_from_file:
        print('OK')
    else:
        print('differs:')
        print(' ', expected)
        print(' ', line_from_file)

def main():
    macros = read_macros_from_file()
    print(f'{SERIALIZE_PATH} has {len(macros)} SERIALIZE* macros')

    index = cindex.Index.create()
    print(f'Parsing {TU_PATH} ...')
    tu = index.parse(TU_PATH)
    checked_kinds = [
        cindex.CursorKind.STRUCT_DECL,
        cindex.CursorKind.CLASS_DECL,
        cindex.CursorKind.CLASS_TEMPLATE,
    ]
    for g1 in tu.cursor.get_children():
        if g1.kind == cindex.CursorKind.NAMESPACE and g1.spelling == 'gemmi':
            for g2 in g1.get_children():
                name = g2.spelling
                if g2.kind in checked_kinds and name in macros:
                    compare_struct(name, g2, macros[name])
                    del macros[name]
                    for g3 in g2.get_children():
                        nested = f'{name}::{g3.spelling}'
                        if g3.kind in checked_kinds and nested in macros:
                            compare_struct(nested, g3, macros[nested])
                            del macros[nested]
    if macros:
        print('\nUNCHECKED:', ' '.join(macros))

if __name__ == '__main__':
    main()



#!/usr/bin/env python

import math
import sys
import unittest
import zlib
import gemmi
from common import full_path, get_path_for_tempfile, assert_numpy_equal, numpy

class TestFloatGrid(unittest.TestCase):
    def test_reading(self):
        m = gemmi.read_ccp4_map(full_path('5i55_tiny.ccp4'))
        self.assertEqual(m.grid.nu, 8)
        self.assertEqual(m.grid.nv, 6)
        self.assertEqual(m.grid.nw, 10)
        self.assertEqual(m.header_i32(28), 0)
        m.set_header_i32(28, 20140)  # set NVERSION
        self.assertEqual(m.header_i32(28), 20140)
        dmax = m.header_float(21)
        self.assertEqual(dmax, max(p.value for p in m.grid))
        self.assertNotEqual(m.grid.axis_order, gemmi.AxisOrder.XYZ)
        m.setup(float('nan'))
        self.assertEqual(m.grid.axis_order, gemmi.AxisOrder.XYZ)
        self.assertEqual(m.grid.nu, 60)
        self.assertEqual(m.grid.nv, 24)
        self.assertEqual(m.grid.nw, 60)
        self.assertEqual(m.grid.point_count, 60 * 24 * 60)
        self.assertEqual(m.header_float(14), 90.0)  # 14 - alpha angle
        self.assertEqual(m.grid.unit_cell.alpha, 90.0)
        self.assertEqual(m.grid.spacegroup.ccp4, 4)  # P21

        extent = m.get_extent()
        self.assertEqual(extent.minimum.tolist(), [-1e-9]*3)
        nu = m.grid.nu
        self.assertTrue((nu - 1.) / nu < extent.maximum.x < 1)

        pos = gemmi.Position(19.4, 3., 21.)
        frac = m.grid.unit_cell.fractionalize(pos)
        pos_value = 2.1543798446655273
        self.assertAlmostEqual(m.grid.interpolate_value(pos), pos_value)
        self.assertAlmostEqual(m.grid.interpolate_value(frac), pos_value)

        # this spacegroup has symop -x, y+1/2, -z
        m.grid.set_value(60-3, 24//2+4, 60-5, 100)  # image of (3, 4, 5)
        self.assertEqual(m.grid.get_value(60-3, 24//2+4, 60-5), 100)
        self.assertTrue(math.isnan(m.grid.get_value(3, 4, 5)))
        m.grid.symmetrize_max()
        self.assertEqual(m.grid.get_value(3, 4, 5), 100)
        m.grid.set_value(3, 4, 5, float('nan'))
        self.assertTrue(math.isnan(m.grid.get_value(3, 4, 5)))
        m.grid.symmetrize_min()
        self.assertEqual(m.grid.get_value(3, 4, 5), 100)
        m.grid.set_value(60-3, 24//2+4, 60-5, 80)
        m.grid.symmetrize_avg()
        self.assertEqual(m.grid.get_value(3, 4, 5), 90)
        m.grid.set_value(60-3, 24//2+4, 60-5, float('nan'))
        m.grid.symmetrize_max()
        self.assertEqual(m.grid.get_value(60-3, 24//2+4, 60-5), 90)
        if numpy:
            arr = m.grid.array
            self.assertEqual(arr.shape, (60, 24, 60))
            self.assertEqual(arr[3][4][5], 90)
            grid2 = gemmi.FloatGrid(arr)
            self.assertTrue(numpy.allclose(m.grid, grid2, atol=0.0, rtol=0,
                                           equal_nan=True))

    def test_new(self):
        N = 24
        m = gemmi.FloatGrid(N, N, N)
        self.assertEqual(m.nu, N)
        self.assertEqual(m.nv, N)
        self.assertEqual(m.nw, N)
        m.set_value(1,2,3, 1.0)
        self.assertEqual(m.sum(), 1.0)
        m.spacegroup = gemmi.find_spacegroup_by_name('C2')
        self.assertEqual(m.spacegroup.number, 5)
        m.symmetrize_max()
        self.assertEqual(m.sum(), 4.0)
        m.get_point(0, 0, 0).value += 1
        self.assertEqual(m.sum(), 5.0)
        m.fill(2.0)
        m.spacegroup = gemmi.find_spacegroup_by_name('P 62 2 2')
        self.assertEqual(len(m.spacegroup.operations()), 12)
        m.set_value(1, 2, 3, 0.0)
        m.symmetrize_min()
        self.assertEqual(m.sum(), 2 * N * N * N - 2 * 12)

    def test_grid_size(self):
        # original cell from 4a0g, and a cell with a <-> b
        cell = gemmi.UnitCell(79.442, 80.066, 136.939, 99.96, 107.12, 97.25)
        cell2 = gemmi.UnitCell(80.066, 79.442, 136.939, 99.96, 107.12, 97.25)
        self.assertAlmostEqual(cell.calculate_d([4, -32, 9]), 2.501956204)
        dmin = 2.5
        self.assertEqual(cell.get_hkl_limits(dmin), [31, 32, 54])
        self.assertEqual(cell2.get_hkl_limits(dmin), [32, 31, 54])
        # In prepare_asu_data(), max_k == (grid.nv - 1) / 2,
        # so for k=31 we need nv>=63, for k=32 we need nv>=65.
        # But when choosing grid size, we take the same n for almost equal
        # lengths (a and b here). So both nu and nv are 72 instead of 64.
        grid = gemmi.FloatGrid()
        grid.spacegroup = gemmi.SpaceGroup('P 1')
        grid.unit_cell = cell
        grid.set_size_from_spacing(dmin / 2, gemmi.GridSizeRounding.Up)
        self.assertTrue([grid.nu, grid.nv, grid.nw] == [72, 72, 120])
        grid.unit_cell = cell2
        grid.set_size_from_spacing(dmin / 2, gemmi.GridSizeRounding.Up)
        self.assertTrue([grid.nu, grid.nv, grid.nw] == [72, 72, 120])

    def test_interpolation(self):
        # Setup grids for testing
        cell = gemmi.UnitCell(10.0,10.0,10.0, 90.0, 90.0, 90.0)
        moving_grid = gemmi.FloatGrid(10, 10, 10)
        moving_grid.spacegroup = gemmi.SpaceGroup('P 1')
        moving_grid.set_unit_cell(cell)
        moving_grid.set_value(0,0,0,1.0)
        moving_grid.set_value(5,5,5,1.0)

        interpolated_grid = gemmi.FloatGrid(10, 10, 10)
        interpolated_grid.spacegroup = gemmi.SpaceGroup('P 1')
        interpolated_grid.set_unit_cell(cell)

        if numpy is None:
            return
        # Test array interpolation
        positions = numpy.array([[0.0,0.0,0.0], [5.0,5.0,5.0]],
                                dtype=numpy.float64)
        values = moving_grid.interpolate_position_array(positions)
        self.assertAlmostEqual(values[0], 1.0)
        self.assertAlmostEqual(values[1], 1.0)

        # Test map morphing
        # A simple single solid translation of the cell, limited to two points
        point_array_list = [
            numpy.array([[1,1,1], [6,6,6]],dtype=numpy.int32),]
        position_array_list = [
            numpy.array([[1.0,1.0,1.0], [6.0,6.0,6.0]],dtype=numpy.float32),]
        ts = [gemmi.Transform(),]
        com_m = [-1.0,-1.0,-1.0,]
        com_r = [0.0,0.0,0.0,]

        ts[0].vec.fromlist(
            (gemmi.Vec3(*com_m) - ts[0].mat.multiply(gemmi.Vec3(*com_r))
             ).tolist()
        )

        interpolated_grid.interpolate_grid_flexible(
            moving_grid,
            point_array_list,
            position_array_list,
            ts,
        )
        self.assertAlmostEqual(interpolated_grid.get_value(1,1,1), 1.0)
        self.assertAlmostEqual(interpolated_grid.get_value(6,6,6), 1.0)


class TestCcp4Map(unittest.TestCase):
    @unittest.skipIf(numpy is None, "NumPy not installed.")
    def test_567_map(self):
        # make a small, contrived map
        data = numpy.arange(5*6*7, dtype=numpy.float32).reshape((5,6,7))
        cell = gemmi.UnitCell(150, 132, 140, 90, 90, 90)
        m = gemmi.Ccp4Map()
        m.grid = gemmi.FloatGrid(data, cell, gemmi.SpaceGroup('P 1'))
        m.update_ccp4_header()

        # write, read and compare
        tmp_path = get_path_for_tempfile(suffix='.ccp4')
        m.write_ccp4_map(tmp_path)
        with open(tmp_path, 'rb') as f:
            tmp_bytes = f.read()
        crc_mod_2_32 = zlib.crc32(tmp_bytes) % 4294967296
        if sys.byteorder == 'little':
            self.assertEqual(crc_mod_2_32, 4078044323)
        elif sys.byteorder == 'big':
            self.assertEqual(crc_mod_2_32, 372922578)
        self.assertEqual(tmp_bytes,
                         m.ccp4_header + m.grid.array.tobytes(order='A'))

        box = gemmi.FractionalBox()
        box.minimum = gemmi.Fractional(0.5/5, 1.5/6, 3.5/7)
        box.maximum = gemmi.Fractional(4.5/5, 2.5/6, 5.5/7)
        m.set_extent(box)
        cut_data = data[1:5, 2:3, 4:6]
        self.assertEqual(cut_data.shape, (4, 1, 2))
        self.assertTrue(numpy.array_equal(m.grid.array, cut_data))

        # CCP4 MAPMASK generated tests/iota_yzx.ccp4.gz from map m:
        #   mapmask mapin iota_full.ccp4 mapout iota_yzx.ccp4 << eof
        #   XYZLIM 0.3 0.9 3.4 3.45 -0.5 -0.3
        #   AXIS Y Z X
        #   MODE mapin
        #   eof
        yzx_path = full_path('iota_yzx.ccp4.gz')
        expanded_data = numpy.full(data.shape, float('nan'), dtype=data.dtype)
        expanded_data[1:5, 2:3, 4:6] = cut_data

        mcut = gemmi.read_ccp4_map(yzx_path, setup=False)
        self.assertEqual(mcut.axis_positions(), [2, 0, 1])
        mcut.setup(float('nan'), gemmi.MapSetup.ReorderOnly)
        self.assertEqual(mcut.axis_positions(), [0, 1, 2])
        self.assertFalse(mcut.full_cell())
        self.assertEqual(mcut.grid.axis_order, gemmi.AxisOrder.Unknown)
        self.assertTrue(numpy.array_equal(mcut.grid.array, cut_data))

        mcut = gemmi.read_ccp4_map(full_path(yzx_path), setup=False)
        self.assertFalse(mcut.full_cell())
        self.assertEqual(mcut.grid.axis_order, gemmi.AxisOrder.Unknown)
        mcut.setup(float('nan'), gemmi.MapSetup.NoSymmetry)
        self.assertTrue(mcut.full_cell())
        self.assertEqual(mcut.grid.axis_order, gemmi.AxisOrder.XYZ)
        assert_numpy_equal(self, mcut.grid.array, expanded_data)

        mcut = gemmi.read_ccp4_map(full_path(yzx_path), setup=True)
        self.assertTrue(mcut.full_cell())
        self.assertEqual(mcut.grid.axis_order, gemmi.AxisOrder.XYZ)
        assert_numpy_equal(self, mcut.grid.array, expanded_data)

    def test_normalize(self):
        yzx_path = full_path('iota_yzx.ccp4.gz')
        m = gemmi.read_ccp4_map(full_path(yzx_path), setup=True)
        m.grid.normalize()
        m.update_ccp4_header()
        mean = m.header_float(22)
        rms = m.header_float(55)
        self.assertAlmostEqual(mean, 0)
        self.assertAlmostEqual(rms, 1)

    @unittest.skipIf(numpy is None, "NumPy not installed.")
    def test_get_subarray(self):
        data = numpy.arange(5*6*7, dtype=numpy.float32).reshape((5,6,7))
        grid = gemmi.FloatGrid(data)
        sub = grid.get_subarray([4,-3,20], [5,10,4])
        self.assertEqual(sub[0][0][0], 195.)
        self.assertEqual(sub[1][2][3], 37.)
        self.assertEqual(sub[-1][-1][-1], 128.)
        grid.set_subarray(-sub, [4,-3,20])
        sub2 = grid.get_subarray([4,-3,20], [5,10,4])
        assert_numpy_equal(self, -sub, sub2)

    @unittest.skipIf(numpy is None, "NumPy not installed.")
    def test_setup_nosymmetry(self):
        m = gemmi.read_ccp4_map(full_path('5i55_tiny.ccp4'))
        orig_point_count = m.grid.point_count
        m.setup(0, gemmi.MapSetup.Full)
        full_nonzero = numpy.count_nonzero(m.grid.array)
        self.assertTrue(full_nonzero > orig_point_count)
        m = gemmi.read_ccp4_map(full_path('5i55_tiny.ccp4'))
        m.setup(0, gemmi.MapSetup.NoSymmetry)
        nosym_nonzero = numpy.count_nonzero(m.grid.array)
        self.assertEqual(full_nonzero, nosym_nonzero * 2)
        nonzero_ext = m.grid.get_nonzero_extent()
        span = nonzero_ext.maximum - nonzero_ext.minimum
        volume = span[0] * span[1] * span[2]
        self.assertAlmostEqual(orig_point_count / m.grid.point_count, volume)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import os
import unittest
import urllib.error
import http.client

import gemmi

# for some reasons, importing common.py here gives me leak warnings
def full_path(filename):
    return os.path.join(os.path.dirname(__file__), filename)

class TestFetch(unittest.TestCase):
    def fetch(self, site):
        cell = gemmi.read_structure(full_path('5i55.cif')).cell
        import gemmi.fetch
        for use_cif in [True, False]:
            try:
                st = gemmi.fetch.read_structure_with_code('5i55', pdb_site=site,
                                                          use_cif=use_cif)
                self.assertEqual(cell, st.cell)
            except urllib.error.URLError as e:
                self.skipTest(f"Cannot connect to {site} ({e})")
            except http.client.RemoteDisconnected as e:
                self.skipTest(f"Couldn't fetch file form {site}: ({e})")
    @unittest.skipIf(__name__ != '__main__', "Skip during unittest discover")
    def test_rcsb(self):
        self.fetch('R')
    @unittest.skipIf(__name__ != '__main__', "Skip during unittest discover")
    def test_pdbe(self):
        self.fetch('E')
    @unittest.skipIf(__name__ != '__main__', "Skip during unittest discover")
    def test_pdbj(self):
        self.fetch('J')

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import unittest
import gemmi
from common import full_path

class TestAlignment(unittest.TestCase):
    def test_string_align(self):
        result = gemmi.align_string_sequences(list('AABCC'),
                                              list('ABC'), [0])
        self.assertEqual(result.score, 0)
        self.assertEqual(result.cigar_str(), '1I3M1I')
        self.assertEqual(result.add_gaps('AABCC', 1), 'AABCC')
        self.assertEqual(result.add_gaps('ABC', 2), '-ABC-')
        self.assertEqual(result.calculate_identity(), 100.)
        self.assertEqual(result.calculate_identity(1), 60.)
        self.assertEqual(result.calculate_identity(2), 100.)
        self.assertEqual(result.match_string, ' ||| ')
        result = gemmi.align_string_sequences(list('SIMILARITY'),
                                              list('PILLAR'), [])
        self.assertEqual(result.match_count, 4)
        self.assertEqual(result.cigar_str(), '3M1I3M3I')

    def test_hemoglobin_alignment(self):
        # based on example from
        # http://biopython.org/DIST/docs/tutorial/Tutorial.html
        hba_human = ("MVLSPADKTNVKAAWGKVGAHAGEYGAEALERMFLSFPTTKTYFPHFDLSHGSAQ"
                     "VKGHGKKVADALTNAVAHVDDMPNALSALSDLHAHKLRVDPVNFKLLSHCLLVTL"
                     "AAHLPAEFTPAVHASLDKFLASVSTVLTSKYR")
        hbb_human = ("MVHLTPEEKSAVTALWGKVNVDEVGGEALGRLLVVYPWTQRFFESFGDLSTPDAV"
                     "MGNPKVKAHGKKVLGAFSDGLAHLDNLKGTFATLSELHCDKLHVDPENFRLLGNV"
                     "LVCVLAHHFGKEFTPPVQAAYQKVVAGVANALAHKYH")
        AA = gemmi.ResidueKind.AA
        hba_seq = gemmi.expand_one_letter_sequence(hba_human, AA)
        hbb_seq = gemmi.expand_one_letter_sequence(hbb_human, AA)
        id_score = gemmi.AlignmentScoring()
        id_score.match = 1
        id_score.mismatch = 0
        id_score.gapo = 0
        id_score.gape = 0
        result = gemmi.align_string_sequences(hba_seq, hbb_seq, [], id_score)
        # "80 different alignments with the score 72"
        self.assertEqual(result.score, 72)
        blosum62 = gemmi.AlignmentScoring('b')
        blosum62.gapo = -9
        result = gemmi.align_string_sequences(hba_seq, hbb_seq, [], blosum62)
        # BioPython equivalent is:
        # pairwise2.align.globalds(seq1.seq, seq2.seq, blosum62, -10, -1)
        self.assertEqual(result.score, 290)

    def test_assign_best_sequences(self):
        st = gemmi.read_structure(full_path('1lzh.pdb.gz'))
        st.setup_entities()
        seq3 = st.entities[0].full_sequence
        self.assertEqual(len(seq3), 129)
        seq1 = gemmi.one_letter_code(seq3)
        self.assertEqual(len(seq1), 129)
        st.clear_sequences()
        self.assertEqual(len(st.entities[0].full_sequence), 0)
        st.assign_best_sequences([seq1+'A', seq1])
        self.assertEqual(len(st.entities[0].full_sequence), 129)
        self.assertEqual(st.entities[0].subchains, ['Axp', 'Bxp'])

    def test_superposition(self):
        model = gemmi.read_structure(full_path('4oz7.pdb'))[0]
        poly1 = model['A'].get_polymer()
        poly2 = model['B'].get_polymer()
        ptype = poly1.check_polymer_type()
        S = gemmi.SupSelect
        s1 = gemmi.calculate_superposition(poly1, poly2, ptype, S.CaP)
        s2 = gemmi.calculate_superposition(poly1, poly2, ptype, S.MainChain)
        s3 = gemmi.calculate_superposition(poly1, poly2, ptype, S.All)
        self.assertEqual(s1.count, 10)
        self.assertEqual(s2.count, 39)
        self.assertEqual(s3.count, 77)
        self.assertAlmostEqual(s1.rmsd, 0.146, places=3)
        self.assertAlmostEqual(s2.rmsd, 0.174, places=3)
        self.assertAlmostEqual(s3.rmsd, 0.400, places=3)
        for s in [s1, s2, s3]:
            self.assertAlmostEqual(s.transform.vec.y, 17.0, places=1)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# Read SF-mmCIF file and plot I/sigma as a function of 1/d^2.

import sys
from matplotlib import pyplot
import gemmi

for path in sys.argv[1:]:
    doc = gemmi.cif.read(path)
    rblock = gemmi.as_refln_blocks(doc)[0]
    intensity = rblock.make_float_array('intensity_meas')
    sigma = rblock.make_float_array('intensity_sigma')
    x = rblock.make_1_d2_array()
    y = intensity / sigma
    pyplot.figure(figsize=(5,5))
    pyplot.hexbin(x, y, gridsize=70, bins='log', cmap='gnuplot2')
    pyplot.show()



#!/usr/bin/env cctbx.python
#
# symmetry.hpp can be checked with:
# python3 tools/gen_sg_table.py $CCP4/lib/data/syminfo.lib > gen.txt
# sed -n '/gen_sg_table/,/^}/p' src/symmetry.cpp > cur.txt
# diff cur.txt gen.txt

import shlex
import sys

from cctbx import sgtbx
try:
    import gemmi
except ImportError:
    gemmi = None

syminfo_path = sys.argv[1]

def parse_syminfo(path):
    data = []
    cur = None
    with open(path) as f:
        for line in f:
            line = line.strip()
            if not line or line[0] == '#':
                continue
            if line == 'begin_spacegroup':
                assert cur is None, line
                cur = {'symops': [], 'cenops': []}
                continue
            assert cur is not None, line
            if line == 'end_spacegroup':
                for must_have in ['basisop', 'ccp4', 'number', 'hall', 'xhm']:
                    assert must_have in cur, must_have
                for must_have_list in ['symops', 'cenops']:
                    assert len(cur[must_have_list]) > 0
                verify_hall_symbol(cur)
                data.append(cur)
                cur = None
            elif line.startswith('number '):
                cur['number'] = int(line[6:])
            elif line.startswith('basisop '):
                cur['basisop'] = line[8:]
            elif line.startswith('symbol ccp4 '):
                cur['ccp4'] = int(line[12:])
                assert cur['ccp4'] == 0 or cur['ccp4'] % 1000 == cur['number']
            elif line.startswith('symbol Hall '):
                cur['hall'] = line[12:].strip(" '")
            elif line.startswith('symbol xHM '):
                cur['xhm'] = line[11:].strip(" '").replace(' :', ':')
            elif line.startswith('symbol old '):
                cur['old'] = shlex.split(line[11:])
            elif line.startswith('symop '):
                cur['symops'].append(line[6:])
            elif line.startswith('cenop '):
                cur['cenops'].append(line[6:])
    return data

def verify_hall_symbol(entry):
    if not gemmi:
        return
    hall_ops = gemmi.symops_from_hall(entry['hall'])
    assert len(hall_ops.sym_ops) == len(entry['symops'])
    assert len(hall_ops.cen_ops) == len(entry['cenops'])
    # centering vectors are exactly the same
    assert (set(gemmi.Op().translated(tr) for tr in hall_ops.cen_ops)
            == set(gemmi.Op(e) for e in entry['cenops'])), entry
    # symops differ in some cases but are the same modulo centering vectors
    given = set(gemmi.Op(s) * c for s in entry['symops']
                for c in entry['cenops'])
    assert given == set(hall_ops), entry

basisops = []

def get_basisop(sgtbx_info):
    to_ref = sgtbx_info.change_of_basis_op_to_reference_setting()
    from_ref = to_ref.inverse()
    b_str = str(from_ref.c())
    try:
        b_idx = basisops.index(b_str)
    except ValueError:
        b_idx = len(basisops)
        basisops.append(b_str)
    #print('FR %s' % from_ref.c())
    #print('TR %s' % to_ref.c())
    return b_str, b_idx

shorter_halls = {
    'C 2y (x+1/4,y+1/4,-x+z-1/4)': 'I 2yb',
    'C 2y (x+1/4,y+1/4,z)': 'C 2yb',
    'P 2 2ab (x+1/4,y+1/4,z)': 'P 2ab 2a',
    'C 2c 2 (x+1/4,y,z)': 'C 2ac 2',
    'C 2 2 (x+1/4,y+1/4,z)': 'C 2ab 2b',
    'F 2 2 (x,y,z+1/4)': 'F 2 2c',
    'I 2 2 (x-1/4,y+1/4,z-1/4)': 'I 2ab 2bc',
    'P 4n 2n (x-1/4,y-1/4,z-1/4)': 'P 4bc 2a',
    'I 2 2 3 (x+1/4,y+1/4,z+1/4)': 'I 2ab 2bc 3',
}

syminfo_data = parse_syminfo(syminfo_path)
syminfo_dict = {}
for info in syminfo_data:
    # handle duplicates in syminfo.lib: pick the the one with ccp4 number
    if info['xhm'] not in syminfo_dict or info['ccp4'] != 0:
        syminfo_dict[info['xhm']] = info


print('''\
  // This table was generated by tools/gen_sg_table.py.
  // First 530 entries in the same order as in SgInfo, sgtbx and ITB.
  // Note: spacegroup 68 has three duplicates with different H-M names.''')

hall_to_idx = {}
counter = 0
def check_dup(hall):
    n = hall_to_idx.setdefault(hall, counter)
    return '' if n == counter else f' (=={n})'
fmt = '  {%3d, %4d, %-12s, %s, %6s, %-16s, %-2d}, // %3d%s'
def quot(s):
    return '"%s"' % s.replace('"', r'\"')
for s in sgtbx.space_group_symbol_iterator():
    xhm = s.hermann_mauguin()
    ext = '  0'
    if s.extension() != '\0':
        xhm += ':%s' % s.extension()
        ext = "'%c'" % s.extension()
    info = syminfo_dict.pop(xhm)
    assert info['number'] == s.number()
    cctbx_sg = sgtbx.space_group(s.hall())
    cctbx_info = sgtbx.space_group_info(group=cctbx_sg)
    from_ref, basisop_idx = get_basisop(cctbx_info)
    assert from_ref == info['basisop'], (from_ref, info['basisop'])
    if gemmi:
        assert (set(gemmi.symops_from_hall(s.hall()))
                == set(gemmi.symops_from_hall(info['hall']))), xhm
    hall = s.hall().strip()
    print(fmt % (s.number(), info['ccp4'], quot(s.hermann_mauguin()),
                 ext, quot(s.qualifier()), quot(hall),
                 basisop_idx, counter, check_dup(hall)))
    counter += 1
print('  // And extra entries from syminfo.lib')
for info in syminfo_data:
    if info['xhm'] in syminfo_dict:  # i.e. it was not printed yet
        hm = info['xhm']
        if not hm:
            hm = info['old'][0]
            if hm == 'P 21 21 2 (a)':  # slightly too long
                hm = 'P 21212(a)'
        ext = '  0'
        if ':' in hm:
            hm, ext = hm.split(':')
            hm = hm.strip()
            ext = "'%c'" % ext
        hall = shorter_halls[info['hall']]
        if gemmi:
            assert (set(gemmi.symops_from_hall(hall))
                    == set(gemmi.symops_from_hall(info['hall']))), hm
        cctbx_sg = sgtbx.space_group(hall)
        cctbx_info = sgtbx.space_group_info(group=cctbx_sg)
        from_ref, basisop_idx = get_basisop(cctbx_info)
        print(fmt % (info['number'], info['ccp4'], quot(hm), ext, '""',
                     quot(hall), basisop_idx, counter, check_dup(hall)))
        counter += 1
print('  // And ...')
additional = [
    'triclinic - enlarged unit cells',
    ('A 1', 'A 1'),
    ('B 1', 'B 1'),
    ('C 1', 'C 1'),
    ('F 1', 'F 1'),
    ('I 1', 'I 1'),
    ('A -1', '-A 1'),
    ('B -1', '-B 1'),
    ('C -1', '-C 1'),
    ('F -1', '-F 1'),
    ('I -1', '-I 1'),
    'monoclinic (qualifiers such as "b1" are assigned arbitrary unique numbers)',  # noqa:E501
    ('B 1 2 1', 'B 2y'),  # 3
    ('C 1 1 2', 'C 2'),  # 3
    ('B 1 21 1', 'B 2yb'),  # 4
    ('C 1 1 21', 'C 2c'),  # 4
    ('F 1 2 1', 'F 2y'),  # 5
    ('F 1 m 1', 'F -2y'),  # 8
    ('F 1 d 1', 'F -2yuw'),  # 9
    ('F 1 2/m 1', '-F 2y'),  # 12
    'orthorhombic',
    ('A b a m', '-A 2 2ab'),
    'tetragonal - enlarged C- and F-centred unit cells',
    ('C 4 2 2', 'C 4 2'),
    ('C 4 2 21', 'C 4a 2'),
    ('F 4 2 2', 'F 4 2'),
    ('C -4 2 m', 'C -4 2'),
    ('C -4 2 b', 'C -4 2ya'),
    ('F 4/m m m', '-F 4 2'),
]

def has_symmorphic(hall):
    symmor_ops = gemmi.symops_from_hall(hall).derive_symmorphic()
    if gemmi.find_spacegroup_by_ops(symmor_ops):
        return True
    return any(gemmi.symops_from_hall(x[1]) == symmor_ops for x in additional
               if type(x) is tuple)

def find_unique_axis(hm):
    s = hm.split()
    assert len(s) == 4
    if s[1] == s[3]:
        return 'b'
    if s[2] == s[3]:
        return 'a'
    if s[1] == s[2]:
        return 'c'
    return '?'

system = None
for item in additional:
    if type(item) is str:
        print('  // ' + item)
        system = item.split()[0]
        continue
    hm, hall = item
    cctbx_sg = sgtbx.space_group(hall)
    cctbx_info = sgtbx.space_group_info(group=cctbx_sg)
    from_ref, basisop_idx = get_basisop(cctbx_info)
    number = cctbx_info.type().number()
    qual = '""'
    if system == 'monoclinic':
        qual = '"%s"' % find_unique_axis(hm)
    print(fmt % (number, 0, quot(hm), '  0', qual, quot(hall),
                 basisop_idx, counter, check_dup(hall)))
    if gemmi:
        assert gemmi.SpaceGroup(number).crystal_system_str() == system
        if not has_symmorphic(hall):
            print('// -- check derive_symmorphic() for', hm)
    counter += 1

print('\n')
for n, b in enumerate(basisops):
    nice_b = gemmi.Op(b).triplet()
    print('    "%s",  // %d' % (nice_b, n))



#!/usr/bin/env python

import unittest
import math
import cmath
import gemmi
from common import full_path, numpy

# from 5nl9
FRAGMENT_WITH_UNK = """\
CRYST1   48.683   56.100   68.794  90.00 105.82  90.00 P 1 21 1      4
ATOM   2320  N   PRO B 145      27.193  12.782  46.813  1.00 43.92           N
ATOM   2321  CA  PRO B 145      27.971  11.557  46.488  1.00 44.82           C
ATOM   2322  C   PRO B 145      27.109  10.394  45.985  1.00 63.82           C
ATOM   2323  O   PRO B 145      25.961  10.269  46.451  1.00 67.23           O
ATOM   2324  CB  PRO B 145      28.599  11.170  47.841  1.00 46.06           C
ATOM   2325  CG  PRO B 145      28.536  12.388  48.681  1.00 49.45           C
ATOM   2326  CD  PRO B 145      27.276  13.086  48.254  1.00 44.74           C
ATOM   2327  OXT PRO B 145      27.597   9.587  45.164  1.00 83.13           O
TER    2328      PRO B 145
HETATM 2329 ZN    ZN A 201       3.818  16.117  25.579  1.00 33.14          ZN2+
HETATM 2330  K     K A 202      -6.950  11.545  40.774  0.80 41.06           K1+
HETATM 2331 ZN    ZN B 200       7.588   6.321  46.448  1.00 40.71          ZN2+
HETATM 2332  UNK UNX B 201       8.947   4.729  45.579  1.00 37.10           X
HETATM 2333  UNK UNX B 202       5.981   5.180  45.153  1.00 39.91           X
HETATM 2334  O   HOH A 301       6.823  17.093  -0.032  1.00 42.64           O
"""

class TestDensityCalculator(unittest.TestCase):
    def test_unk_does_not_cause_segfault(self):
        st = gemmi.read_pdb_string(FRAGMENT_WITH_UNK)
        for calculator in [gemmi.DensityCalculatorX,
                           gemmi.DensityCalculatorE,
                           gemmi.DensityCalculatorN]:
            dencalc = calculator()
            dencalc.d_min = 2.1
            dencalc.grid.setup_from(st)
            # we only check here that it doesn't crash
            dencalc.put_model_density_on_grid(st[0])

class TestExpandingToP1(unittest.TestCase):
    def test_1gdr(self):
        st = gemmi.read_pdb(full_path('pdb1gdr.ent'), max_line_length=72)
        sg = st.find_spacegroup()
        order = len(sg.operations())
        self.assertEqual(order, 12)
        sfcalc = gemmi.StructureFactorCalculatorX(st.cell)
        mtz = gemmi.Mtz(with_base=True)
        mtz.spacegroup = sg
        mtz.set_cell_for_all(st.cell)
        mtz.add_dataset('calc')
        mtz.add_column('FC', 'F')
        mtz.add_column('PHIC', 'P')
        hkl1 = (-7, 11, 5)
        hkl2 = (6, 8, -1)
        sf1 = sfcalc.calculate_sf_from_model(st[0], hkl1)
        sf2 = sfcalc.calculate_sf_from_model(st[0], hkl2)
        def polar(c):
            mag, phase = cmath.polar(c)
            return (mag, math.degrees(phase))
        if not numpy:
            return
        data = numpy.array([hkl1 + polar(sf1),
                            hkl2 + polar(sf2)], numpy.float32)
        mtz.set_data(data)
        mtz.expand_to_p1()
        st.transform_to_assembly('unit_cell',
                                 gemmi.HowToNameCopiedChain.AddNumber)
        # AsuData as_is is not yet in ASU.
        pre_asudata = mtz.get_f_phi('FC', 'PHIC', as_is=True)
        # Check structure factors after expand_to_p1(),
        # in particular we check here if phase shift is correct.
        for v in pre_asudata:
            usf = sfcalc.calculate_sf_from_model(st[0], v.hkl)
            self.assertAlmostEqual(v.value, usf / order, delta=1e-4)
        # The same after Mtz.ensure_asu(), which also changes phases.
        mtz.ensure_asu()
        asudata = mtz.get_f_phi('FC', 'PHIC', as_is=True)
        self.assertEqual(hkl1, tuple(asudata[0].hkl))
        for v in asudata:
            usf = sfcalc.calculate_sf_from_model(st[0], v.hkl)
            self.assertAlmostEqual(v.value, usf / order, delta=1e-4)
        # Finally, test AsuData.ensure_asu().
        pre_asudata.ensure_asu()
        for v in pre_asudata:
            usf = sfcalc.calculate_sf_from_model(st[0], v.hkl)
            self.assertAlmostEqual(v.value, usf / order, delta=1e-4)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# Show differences between two Refmac intermediate files.
# The intermediate files should have content of both crd and rst files.
# Usage: ./crd-diff.py first.crd second.crd

import argparse
from collections import namedtuple
from itertools import zip_longest
from gemmi import cif

COLUMNS = ['record', 'number', 'label', 'period',
           'atom_id_1', 'atom_id_2', 'atom_id_3', 'atom_id_4',
           'value', 'dev', 'val_obs']

Restraint = namedtuple('Restraint', COLUMNS)
Crd = namedtuple('Crd', ['atoms', 'real_serial'])
ignore_tors = False

def read_rst(doc):
    result = []
    for row in doc['restraints'].find('_restr.', COLUMNS):
        if row[0] in ['MONO', 'LINK']:
            result.append((row[0].lower(), row[2].lower(), []))
        else:
            data = Restraint(*[x.lower() if x != '.' else None for x in row])
            if ignore_tors and data.record == 'tors':
                continue
            result[-1][2].append(data)
    return [r for r in result if r[2]]

def read_crd(doc):
    block = doc[0]
    sites = block.find('_atom_site.', ['id', 'label_atom_id', 'label_alt_id',
                                       'label_comp_id', 'occupancy',
                                       'calc_flag'])
    atoms = [a for a in sites if a[-1] != 'M']
    len_diff = len(sites) - len(atoms)
    if len_diff != 0:
        print('Removed %d atoms with calc_flag=M' % len_diff)
    real_serial = {None: None, '.': '.'}
    for a in atoms:
        real_serial[a[0]] = len(real_serial) - 1
    return Crd(atoms, real_serial)

def get_crd_atom_row(crd, atom_id):
    idx = crd.real_serial[atom_id] - 1
    return crd.atoms[idx]

# hydrogen distances in Refmac may not be ideal
def can_have_wrong_val_obs(crd, rst1):
    def is_hydrogen(atom_id):
        atom = get_crd_atom_row(crd, atom_id)
        return cif.as_string(atom[1]).startswith('H')
    if rst1.record == 'bond':
        return is_hydrogen(rst1.atom_id_2)
    elif rst1.record == 'angl':
        return is_hydrogen(rst1.atom_id_3)
    elif rst1.record == 'tors':
        return is_hydrogen(rst1.atom_id_4)
    elif rst1.record == 'plan':
        # we can't easily check here if there are H's in the plane
        return True
        #return is_hydrogen(rst1.atom_id_1)
    return False

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-v', action='store_true', help='verbose output')
    parser.add_argument('--no-tors', action='store_true',
                        help='ignore torsion angles')
    parser.add_argument('file1_crd', metavar='file1.crd')
    parser.add_argument('file2_crd', metavar='file2.crd')
    args = parser.parse_args()
    if args.no_tors:
        global ignore_tors
        ignore_tors = True

    doc1 = cif.read(args.file1_crd)
    doc2 = cif.read(args.file2_crd)

    crd1 = read_crd(doc1)
    crd2 = read_crd(doc2)
    if len(crd1.atoms) != len(crd2.atoms):
        print('_atom_site count differs: %d vs %d' %
              (len(crd1.atoms), len(crd2.atoms)))
    for a1, a2 in zip(crd1.atoms, crd2.atoms):
        if any(a1.str(i) != a2.str(i) for i in [1, 2, 3, 5]):
            print('First difference:')
            print('ATOM %s %s %s %s' % (a1[0], a1.str(1), a1[2], a1[3]))
            print('ATOM %s %s %s %s' % (a2[0], a2.str(1), a2[2], a2[3]))
            return
        if float(a1[4]) != float(a2[4]):
            print('ATOM %s/%s %s %s %s occupancy %s vs %s' % (
                  a1[0], a2[0], a1.str(1), a1[2], a1[3], a1[4], a2[4]))

    r1 = read_rst(doc1)
    r2 = read_rst(doc2)
    mono_count_1 = sum(t[0] == 'mono' for t in r1)
    mono_count_2 = sum(t[0] == 'mono' for t in r2)
    if mono_count_1 == mono_count_2:
        print('MONO count same:', mono_count_1)
    else:
        print('MONO count differs:', mono_count_1, mono_count_2)
    link_count_1 = sum(t[0] == 'link' for t in r1)
    link_count_2 = sum(t[0] == 'link' for t in r2)
    if link_count_1 == link_count_2:
        print('LINK count same:', link_count_1)
    else:
        print('LINK count differs:', link_count_1, link_count_2)

    def same_nums(a, b, eps=0.003, mod360=False):
        if a == b:
            return True
        abs_diff = abs(float(a) - float(b))
        return abs_diff < eps or (mod360 and abs(abs_diff - 360) < eps)

    def fmt_atom_id(atom_id, crd):
        if atom_id is None:
            return '.'
        atom = get_crd_atom_row(crd, atom_id)
        alt = ''
        if atom[2] != '.':
            alt = '.' + atom[2]
        return '%s(%s %s%s)' % (atom_id, atom[3], atom[1], alt)

    def fmt(r, crd):
        return '%s %s %s %s  %s : %s : %s : %s   %s ± %s  (%s)' % (
               r.record, r.number, r.label or '.', r.period or '.',
               fmt_atom_id(r.atom_id_1, crd), fmt_atom_id(r.atom_id_2, crd),
               fmt_atom_id(r.atom_id_3, crd), fmt_atom_id(r.atom_id_4, crd),
               r.value, r.dev, r.val_obs)

    def has_same_ids(rst1, rst2):
        return (rst1.record == rst2.record
                and all(crd1.real_serial[u] == crd2.real_serial[v]
                        for u, v in zip(rst1[4:8], rst2[4:8])))

    for n, (a, b) in enumerate(zip_longest(r1, r2, fillvalue=[None, None, []])):
        if a[:2] != b[:2]:
            print('item', n, 'differs:', a[:2], 'vs', b[:2])
            if a[0] != b[0]:
                print('stop')
                break
        else:
            if args.v:
                print('---', n, a[0], a[1])
        if len(a[2]) != len(b[2]):
            print('Item %d (%s) has different restr count: %d vs %d' %
                  (n, a[0], len(a[2]), len(b[2])))
        n_rst1 = 0
        n_rst2 = 0
        while n_rst1 < len(a[2]) or n_rst2 < len(b[2]):
            r_str = '%s/%s restraint %d:%d/%d' % (b[0], b[1], n, n_rst1, n_rst2)
            if n_rst1 >= len(a[2]):
                rst2 = b[2][n_rst2]
                print('ADDED record %s:\n%s\n' % (r_str, fmt(rst2, crd2)))
                n_rst2 += 1
                continue
            if n_rst2 >= len(b[2]):
                rst1 = a[2][n_rst1]
                print('REMOVED record %s:\n%s\n' % (r_str, fmt(rst1, crd1)))
                n_rst1 += 1
                continue
            rst1 = a[2][n_rst1]
            rst2 = b[2][n_rst2]
            same_ids = has_same_ids(rst1, rst2)
            if not same_ids:
                a2 = a[2]
                b2 = b[2]
                if any(n_rst2+k < len(b2) and has_same_ids(rst1, b2[n_rst2+k])
                       for k in range(1,10)):
                    print('ADDED record %s:\n%s\n' % (r_str, fmt(rst2, crd2)))
                    n_rst2 += 1
                    continue
                if any(n_rst1+k < len(a2) and has_same_ids(a2[n_rst1+k], rst2)
                       for k in range(1,10)):
                    print('REMOVED record %s:\n%s\n' % (r_str, fmt(rst1, crd1)))
                    n_rst1 += 1
                    continue
            if rst1.record != rst2.record:
                print('Different records %s:\n%s\nvs\n%s\n' %
                      (r_str, fmt(rst1, crd1), fmt(rst2, crd2)))
            elif (rst1.label != rst2.label if rst1.record != 'chir'
                  else rst1.label[0] != rst2.label[0]):
                print('Different labels in %s:\n%s\nvs\n%s\n' %
                      (r_str, fmt(rst1, crd1), fmt(rst2, crd2)))
            elif rst1.period != rst2.period:
                print('Different period in %s:\n%s\nvs\n%s\n' %
                      (r_str, fmt(rst1, crd1), fmt(rst2, crd2)))
            elif not same_ids:
                print('Different atom id in %s:\n%s\nvs\n%s\n' %
                      (r_str, fmt(rst1, crd1), fmt(rst2, crd2)))
            elif not same_nums(rst1.value, rst2.value):
                print('Different value for %s (%s vs %s) in:\n%s\n' %
                      (r_str, rst1.value, rst2.value, fmt(rst1, crd1)))
            elif not same_nums(rst1.dev, rst2.dev):
                print('Different dev for %s (%s vs %s) in:\n%s\n' %
                      (r_str, rst1.dev, rst2.dev, fmt(rst1, crd1)))
            elif not (same_nums(rst1.val_obs, rst2.val_obs,
                                eps=val_obs_eps(rst1.record),
                                mod360=(rst1.record == 'tors'))
                      or can_have_wrong_val_obs(crd1, rst1)):
                print('val_obs differs for %s (%s vs %s) in:\n%s\n' %
                      (r_str, rst1.val_obs, rst2.val_obs, fmt(rst1, crd1)))
            #elif rst1.number != rst2.number:
            #    print('Serial number differs in %s: %s -> %s' %
            #          (r_str, rst1[1], rst2[1]))
            n_rst1 += 1
            n_rst2 += 1

def val_obs_eps(record):
    if record == 'tors':
        return 0.15
    elif record == 'plan':
        return 0.015
    else:
        return 0.004

if __name__ == '__main__':
    try:
        main()
    except BrokenPipeError:
        pass
